"""
===============================================================================
883a3cd0218b509abbb32ef169503f1e
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import AsyncCogniteClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes.principals import Principal, PrincipalList
from cognite.client.utils._async_helpers import run_sync
from cognite.client.utils.useful_types import SequenceNotStr


class SyncPrincipalsAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client

    def me(self) -> Principal:
        """
        `Get the current caller's information. <https://developer.cognite.com/api#tag/Principals/operation/getMe>`_

        Returns:
            Principal: The principal of the user running the code, i.e. the
                principal *this* AsyncCogniteClient was instantiated with.

        Examples:
            Get your own principal:
                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.iam.principals.me()
        """
        return run_sync(self.__async_client.iam.principals.me())

    @overload
    def retrieve(self, id: str, external_id: None = None, ignore_unknown_ids: bool = False) -> Principal | None: ...

    @overload
    def retrieve(self, id: None = None, *, external_id: str, ignore_unknown_ids: bool = False) -> Principal | None: ...

    @overload
    def retrieve(
        self, id: SequenceNotStr[str], external_id: None = None, ignore_unknown_ids: bool = False
    ) -> PrincipalList: ...

    @overload
    def retrieve(
        self, id: None = None, *, external_id: SequenceNotStr[str], ignore_unknown_ids: bool = False
    ) -> PrincipalList: ...

    @overload
    def retrieve(
        self, id: None = None, external_id: None = None, ignore_unknown_ids: bool = False
    ) -> PrincipalList: ...

    def retrieve(
        self,
        id: str | SequenceNotStr[str] | None = None,
        external_id: str | SequenceNotStr[str] | None = None,
        ignore_unknown_ids: bool = False,
    ) -> Principal | PrincipalList | None:
        """
        `Retrieve principal by reference in the organization <https://developer.cognite.com/api#tag/Principals/operation/getPrincipalsById>`_

        Args:
            id (str | SequenceNotStr[str] | None): The ID(s) of the principal(s) to retrieve.
            external_id (str | SequenceNotStr[str] | None): The external ID(s) of the principal to retrieve.
            ignore_unknown_ids (bool): This is only relevant when retrieving multiple principals. If set to True, the method will return the principals that were found and ignore the ones that were not found. If set to False, the method will raise a CogniteAPIError if any of the specified principals were not found. Defaults to False.

        Returns:
            Principal | PrincipalList | None: The principal(s) with the specified ID(s) or external ID(s).

        Examples:
            Retrieve a principal by ID:
                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.iam.principals.retrieve(id="20u3of8-1234-5678-90ab-cdef12345678")

            Retrieve a principal by external ID:
                >>> res = client.iam.principals.retrieve(external_id="my_external_id")
        """
        return run_sync(
            self.__async_client.iam.principals.retrieve(
                id=id,  # type: ignore [arg-type]
                external_id=external_id,  # type: ignore [arg-type]
                ignore_unknown_ids=ignore_unknown_ids,
            )
        )

    def list(self, types: str | Sequence[str] | None = None, limit: int = DEFAULT_LIMIT_READ) -> PrincipalList:
        """
        `List principals in the organization <https://developer.cognite.com/api#tag/Principals/operation/listPrincipals>`_

        Args:
            types (str | Sequence[str] | None): Filter by principal type(s). Defaults to None, which means no filtering.
            limit (int): The maximum number of principals to return. Defaults to 25.

        Returns:
            PrincipalList: The principal of the user running the code, i.e. the principal *this* CogniteClient was instantiated with.

        Examples:
            List principals in the organization:
                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.iam.principals.list(types="USER", limit=10)
        """
        return run_sync(self.__async_client.iam.principals.list(types=types, limit=limit))



from __future__ import annotations

from collections.abc import AsyncIterator, Sequence
from typing import TYPE_CHECKING, overload

from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.simulators.filters import SimulatorIntegrationFilter
from cognite.client.data_classes.simulators.simulators import SimulatorIntegration, SimulatorIntegrationList
from cognite.client.utils._experimental import FeaturePreviewWarning
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient
    from cognite.client.config import ClientConfig


class SimulatorIntegrationsAPI(APIClient):
    _RESOURCE_PATH = "/simulators/integrations"

    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: AsyncCogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self._DELETE_LIMIT = 1
        self._warning = FeaturePreviewWarning(
            api_maturity="General Availability", sdk_maturity="alpha", feature_name="Simulators"
        )

    @overload
    def __call__(
        self,
        chunk_size: int,
        simulator_external_ids: str | SequenceNotStr[str] | None = None,
        active: bool | None = None,
        limit: int | None = None,
    ) -> AsyncIterator[SimulatorIntegrationList]: ...

    @overload
    def __call__(
        self,
        chunk_size: None = None,
        simulator_external_ids: str | SequenceNotStr[str] | None = None,
        active: bool | None = None,
        limit: int | None = None,
    ) -> AsyncIterator[SimulatorIntegration]: ...

    async def __call__(
        self,
        chunk_size: int | None = None,
        simulator_external_ids: str | SequenceNotStr[str] | None = None,
        active: bool | None = None,
        limit: int | None = None,
    ) -> AsyncIterator[SimulatorIntegration] | AsyncIterator[SimulatorIntegrationList]:
        """Iterate over simulator integrations

        Fetches simulator integrations as they are iterated over, so you keep a limited number of simulator integrations in memory.

        Args:
            chunk_size (int | None): Number of simulator integrations to return in each chunk. Defaults to yielding one simulator integration a time.
            simulator_external_ids (str | SequenceNotStr[str] | None): Filter on simulator external ids.
            active (bool | None): Filter on active status of the simulator integration.
            limit (int | None): The maximum number of simulator integrations to return, pass None to return all.

        Yields:
            SimulatorIntegration | SimulatorIntegrationList: yields SimulatorIntegration one by one if chunk_size is not specified, else SimulatorIntegrationList objects.
        """  # noqa: DOC404
        integrations_filter = SimulatorIntegrationFilter(simulator_external_ids=simulator_external_ids, active=active)
        async for item in self._list_generator(
            list_cls=SimulatorIntegrationList,
            resource_cls=SimulatorIntegration,
            method="POST",
            filter=integrations_filter.dump(),
            chunk_size=chunk_size,
            limit=limit,
        ):
            yield item

    async def list(
        self,
        limit: int | None = DEFAULT_LIMIT_READ,
        simulator_external_ids: str | SequenceNotStr[str] | None = None,
        active: bool | None = None,
    ) -> SimulatorIntegrationList:
        """`Filter simulator integrations <https://developer.cognite.com/api#tag/Simulator-Integrations/operation/filter_simulator_integrations_simulators_integrations_list_post>`_

        Retrieves a list of simulator integrations that match the given criteria.

        Args:
            limit (int | None): The maximum number of simulator integrations to return, pass None to return all.
            simulator_external_ids (str | SequenceNotStr[str] | None): Filter on simulator external ids.
            active (bool | None): Filter on active status of the simulator integration.

        Returns:
            SimulatorIntegrationList: List of simulator integrations

        Examples:
            List a few simulator integrations:
                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.simulators.integrations.list(limit=10)

            Iterate over simulator integrations, one-by-one:
                >>> for integration in client.simulators.integrations():
                ...     integration  # do something with the simulator integration

            Filter simulator integrations by simulator external ids and active status:
                >>> res = client.simulators.integrations.list(
                ...     simulator_external_ids=["sim1", "sim2"],
                ...     active=True,
                ... )
        """
        integrations_filter = SimulatorIntegrationFilter(simulator_external_ids=simulator_external_ids, active=active)
        self._warning.warn()
        return await self._list(
            method="POST",
            limit=limit,
            resource_cls=SimulatorIntegration,
            list_cls=SimulatorIntegrationList,
            filter=integrations_filter.dump(),
        )

    async def delete(
        self,
        ids: int | Sequence[int] | None = None,
        external_ids: str | SequenceNotStr[str] | None = None,
    ) -> None:
        """`Delete simulator integrations <https://developer.cognite.com/api#tag/Simulator-Integrations/operation/delete_simulator_integrations_simulators_integrations_delete_post>`_

        Args:
            ids (int | Sequence[int] | None): Id(s) of simulator integrations to delete
            external_ids (str | SequenceNotStr[str] | None): External_id(s) of simulator integrations to delete

        Examples:
            Delete simulator integrations by id or external id:
                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> client.simulators.integrations.delete(ids=[1,2,3], external_ids="foo")
        """
        await self._delete_multiple(
            identifiers=IdentifierSequence.load(ids=ids, external_ids=external_ids),
            wrap_ids=True,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.limits import Limit, LimitList
from cognite.client.utils._experimental import FeaturePreviewWarning
from cognite.client.utils._identifier import LimitId

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient
    from cognite.client.config import ClientConfig
    from cognite.client.data_classes.filters import Prefix


class LimitsAPI(APIClient):
    _RESOURCE_PATH = "/limits/values"

    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: AsyncCogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self._warning = FeaturePreviewWarning(
            api_maturity="alpha",
            sdk_maturity="alpha",
            feature_name="Limits API",
        )

    async def retrieve(self, id: str) -> Limit | None:
        """`Retrieve a limit value by its id. <https://api-docs.cognite.com/20230101-alpha/tag/Limits/operation/fetchLimitById/>`_

        Retrieves a limit value by its `limitId`.

        Args:
            id (str): Limit ID to retrieve.
                Limits are identified by an id containing the service name and a service-scoped limit name.
                For instance `atlas.monthly_ai_tokens` is the id of the `atlas` service limit `monthly_ai_tokens`.
                Service and limit names are always in `lower_snake_case`.

        Returns:
            Limit | None: The requested limit, or `None` if not found.

        Examples:

            Retrieve a single limit by id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.limits.retrieve(id="atlas.monthly_ai_tokens")
        """
        self._warning.warn()

        headers = {"cdf-version": f"{self._config.api_subversion}-alpha"}

        return await self._retrieve(
            identifier=LimitId(id),
            cls=Limit,
            headers=headers,
        )

    async def list(self, filter: Prefix | None = None, limit: int | None = DEFAULT_LIMIT_READ) -> LimitList:
        """`List all limit values <https://api-docs.cognite.com/20230101-alpha/tag/Limits/operation/listLimits/>`_

        Retrieves all limit values for a specific project. Optionally filter by limit ID prefix using a `Prefix` filter.

        Args:
            filter (Prefix | None): Optional `Prefix` filter to apply on the `limitId` property (only `Prefix` filters are supported).
            limit (int | None): Maximum number of limits to return. Defaults to 25. Set to None or -1 to return all limits

        Returns:
            LimitList: List of all limit values in the project.

        Examples:

            List all limits:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> limits = client.limits.list(limit=None)

            List limits filtered by prefix (e.g., all limits for the 'atlas' service):

                >>> from cognite.client.data_classes.filters import Prefix
                >>> prefix_filter = Prefix("limitId", "atlas.")
                >>> limits = client.limits.list(filter=prefix_filter)
        """
        self._warning.warn()

        headers = {"cdf-version": f"{self._config.api_subversion}-alpha"}

        return await self._list(
            method="GET" if filter is None else "POST",
            list_cls=LimitList,
            resource_cls=Limit,
            limit=limit,
            filter=None if filter is None else filter.dump(camel_case_property=True),
            headers=headers,
        )



from __future__ import annotations

from abc import ABC
from typing import TYPE_CHECKING, Any, cast

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteFilter,
    CogniteResourceList,
    InternalIdTransformerMixin,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.utils._auxiliary import exactly_one_is_not_none

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


class TransformationNotificationCore(WriteableCogniteResource["TransformationNotificationWrite"], ABC):
    """The transformation notification resource allows configuring email alerts on events related to a transformation run.

    Args:
        destination (str): Email address where notifications should be sent.
    """

    def __init__(
        self,
        destination: str,
    ) -> None:
        self.destination = destination


class TransformationNotification(TransformationNotificationCore):
    """The transformation notification resource allows configuring email alerts on events related to a transformation run.
    This is the read format of a transformation notification.

    Args:
        id (int): A server-generated ID for the object.
        transformation_id (int): Transformation Id.
        transformation_external_id (str): Transformation external Id.
        destination (str): Email address where notifications should be sent.
        created_time (int): Time when the notification was created.
        last_updated_time (int): Time when the notification was last updated.
        cognite_client (AsyncCogniteClient | None): The client to associate with this object.
    """

    def __init__(
        self,
        id: int,
        transformation_id: int,
        transformation_external_id: str,
        destination: str,
        created_time: int,
        last_updated_time: int,
        cognite_client: AsyncCogniteClient | None = None,
    ) -> None:
        super().__init__(destination)
        self.id: int = id
        self.transformation_id: int = transformation_id
        self.transformation_external_id: str = transformation_external_id
        self.created_time: int = created_time
        self.last_updated_time: int = last_updated_time
        self._cognite_client = cast("AsyncCogniteClient", cognite_client)

    def __hash__(self) -> int:
        return hash(self.id)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            id=resource["id"],
            transformation_id=resource["transformationId"],
            transformation_external_id=resource["transformationExternalId"],
            destination=resource["destination"],
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
            cognite_client=cognite_client,
        )

    def as_write(self) -> TransformationNotificationWrite:
        """Returns this Asset in its writing format."""
        if self.destination is None:
            raise ValueError("The write format requires destination to be set")
        return TransformationNotificationWrite(
            destination=self.destination,
            transformation_id=self.transformation_id,
        )


class TransformationNotificationWrite(TransformationNotificationCore):
    """The transformation notification resource allows configuring email alerts on events related to a transformation run.
    This is the write format of a transformation notification.

    Args:
        destination (str): Email address where notifications should be sent.
        transformation_id (int | None): Transformation ID.
        transformation_external_id (str | None): Transformation external ID.
    """

    def __init__(
        self,
        destination: str,
        transformation_id: int | None = None,
        transformation_external_id: str | None = None,
    ) -> None:
        super().__init__(destination)
        if not exactly_one_is_not_none(transformation_id, transformation_external_id):
            raise ValueError("exactly one of transformation_id, transformation_external_id must be specified")
        self.transformation_id = transformation_id
        self.transformation_external_id = transformation_external_id

    def as_write(self) -> TransformationNotificationWrite:
        """Returns this TransformationNotification instance"""
        return self

    @classmethod
    def _load(
        cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None
    ) -> TransformationNotificationWrite:
        return cls(
            destination=resource["destination"],
            transformation_id=resource.get("transformationId"),
            transformation_external_id=resource.get("transformationExternalId"),
        )


class TransformationNotificationWriteList(CogniteResourceList[TransformationNotificationWrite]):
    _RESOURCE = TransformationNotificationWrite


class TransformationNotificationList(
    WriteableCogniteResourceList[TransformationNotificationWrite, TransformationNotification],
    InternalIdTransformerMixin,
):
    _RESOURCE = TransformationNotification

    def as_write(self) -> TransformationNotificationWriteList:
        """Returns this TransformationNotificationList instance"""
        return TransformationNotificationWriteList(
            [item.as_write() for item in self.data], cognite_client=self._get_cognite_client()
        )


class TransformationNotificationFilter(CogniteFilter):
    """TransformationNotificationFilter

    Args:
        transformation_id (int | None): Filter by transformation internal numeric ID.
        transformation_external_id (str | None): Filter by transformation externalId.
        destination (str | None): Filter by notification destination.
    """

    def __init__(
        self,
        transformation_id: int | None = None,
        transformation_external_id: str | None = None,
        destination: str | None = None,
    ) -> None:
        self.transformation_id = transformation_id
        self.transformation_external_id = transformation_external_id
        self.destination = destination



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import UserList
from collections.abc import Collection, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, ClassVar, Literal, TypeAlias, cast, final

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResource,
    CogniteResourceList,
    ExternalIDTransformerMixin,
    InternalIdTransformerMixin,
    UnknownCogniteObject,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.data_modeling.query import Query, ResultSetExpression, Select
from cognite.client.data_classes.simulators.runs import (
    SimulationInputOverride,
)
from cognite.client.utils._experimental import FeaturePreviewWarning
from cognite.client.utils._text import convert_all_keys_to_camel_case, to_snake_case

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient
from zoneinfo import ZoneInfo

TaskStatus: TypeAlias = Literal[
    "in_progress",
    "cancelled",
    "failed",
    "failed_with_terminal_error",
    "completed",
    "completed_with_errors",
    "timed_out",
    "skipped",
]

WorkflowStatus: TypeAlias = Literal["completed", "failed", "running", "terminated", "timed_out"]


class WorkflowCore(WriteableCogniteResource["WorkflowUpsert"], ABC):
    def __init__(self, external_id: str, description: str | None = None, data_set_id: int | None = None) -> None:
        self.external_id = external_id
        self.description = description
        self.data_set_id = data_set_id


class WorkflowUpsert(WorkflowCore):
    """
    This class represents a workflow. This is the write version, used when creating or updating a workflow.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        description (str | None): Description of the workflow. Note that when updating a workflow, the description will
                            always be overwritten also if it is set to None. Meaning if the workflow already has a description,
                            and you want to keep it, you need to provide the description when updating the workflow.
        data_set_id (int | None): The id of the data set this workflow belongs to.
                            If a dataSetId is provided, any operations on this workflow, or its versions, executions,
                            and triggers will require appropriate access to the data set. More information on data sets
                            and their configuration can be found here: https://docs.cognite.com/cdf/data_governance/concepts/datasets/
    """

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            external_id=resource["externalId"],
            description=resource.get("description"),
            data_set_id=resource.get("dataSetId"),
        )

    def as_write(self) -> WorkflowUpsert:
        """Returns this workflow instance."""
        return self


class Workflow(WorkflowCore):
    """
    This class represents a workflow. This is the reading version, used when reading or listing workflows.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        created_time (int): The time when the workflow was created. Unix timestamp in milliseconds.
        last_updated_time (int): The time when the workflow was last updated. Unix timestamp in milliseconds.
        description (str | None): Description of the workflow. Defaults to None.
        data_set_id (int | None): The id of the data set this workflow belongs to.
    """

    def __init__(
        self,
        external_id: str,
        created_time: int,
        last_updated_time: int,
        description: str | None = None,
        data_set_id: int | None = None,
    ) -> None:
        super().__init__(external_id, description, data_set_id)
        self.created_time = created_time
        self.last_updated_time = last_updated_time

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            external_id=resource["externalId"],
            description=resource.get("description"),
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
            data_set_id=resource.get("dataSetId"),
        )

    def as_write(self) -> WorkflowUpsert:
        """Returns this workflow in the writing format."""
        return WorkflowUpsert(
            external_id=self.external_id,
            description=self.description,
            data_set_id=self.data_set_id,
        )


class WorkflowUpsertList(CogniteResourceList[WorkflowUpsert], ExternalIDTransformerMixin):
    _RESOURCE = WorkflowUpsert


class WorkflowList(WriteableCogniteResourceList[WorkflowUpsert, Workflow], ExternalIDTransformerMixin):
    """This class represents a list of workflows."""

    _RESOURCE = Workflow

    def as_write(self) -> WorkflowUpsertList:
        """Returns these workflows in the writing format."""
        return WorkflowUpsertList([workflow.as_write() for workflow in self.data])


ValidTaskType = Literal["function", "transformation", "cdf", "dynamic", "subworkflow", "simulation"]


class WorkflowTaskParameters(CogniteObject, ABC):
    task_type: ClassVar[ValidTaskType]

    @classmethod
    def load_parameters(cls, data: dict) -> WorkflowTaskParameters:
        type_ = data.get("type", data.get("taskType"))
        parameters = data.get("parameters", data.get("input"))
        if parameters is None:
            raise ValueError(
                "You must provide parameter data either with key "
                "'parameter' or 'input', with parameter taking precedence."
            )

        if type_ == "function":
            return FunctionTaskParameters._load(parameters)
        elif type_ == "transformation":
            return TransformationTaskParameters._load(parameters)
        elif type_ == "cdf":
            return CDFTaskParameters._load(parameters)
        elif type_ == "dynamic":
            return DynamicTaskParameters._load(parameters)
        elif type_ == "subworkflow" and "tasks" in parameters["subworkflow"]:
            return SubworkflowTaskParameters._load(parameters)
        elif type_ == "subworkflow" and "workflowExternalId" in parameters["subworkflow"]:
            return SubworkflowReferenceParameters._load(parameters)
        elif type_ == "simulation":
            return SimulationTaskParameters._load(parameters)
        else:
            raise ValueError(f"Unknown task type: {type_}. Expected {ValidTaskType}")


class FunctionTaskParameters(WorkflowTaskParameters):
    """The function parameters are used to specify the Cognite Function to be called.

    Args:
        external_id (str): The external ID of the function to be called.
        data (dict | str | None): The data to be passed to the function. Defaults to None. The data can be used to specify the input to the function from previous tasks or the workflow input. See the tip below for more information.
        is_async_complete (bool | None): Whether the function is asynchronous. Defaults to None, which the API will interpret as False.

    If a function is asynchronous, you need to call the client.workflows.tasks.update() endpoint to update the status of the task.
    While synchronous tasks update the status automatically.

    Tip:
        You can dynamically specify data from other tasks or the workflow. You do this by following the format
        `${prefix.jsonPath}` in the expression. The valid are:

        - `${workflow.input}`: The workflow input.
        - `${<taskExternalId>.output}`: The output of the task with the given external id.
        - `${<taskExternalId>.input}`: The input of the task with the given external id.
        - `${<taskExternalId>.input.someKey}`: A specific key within the input of the task with the given external id.

        For example, if you have a workflow containing two tasks, and the external_id of the first task is `task1` then,
        you can specify the data for the second task as follows:

            >>> from cognite.client.data_classes  import WorkflowTask, FunctionTaskParameters
            >>> task = WorkflowTask(
            ...     external_id="task2",
            ...     parameters=FunctionTaskParameters(
            ...         external_id="cdf_deployed_function",
            ...         data={
            ...             "workflow_data": "${workflow.input}",
            ...             "task1_input": "${task1.input}",
            ...             "task1_output": "${task1.output}"
            ...         },
            ...     ),
            ... )
    """

    task_type = "function"

    def __init__(
        self,
        external_id: str,
        data: dict | str | None = None,
        is_async_complete: bool | None = None,
    ) -> None:
        self.external_id = external_id
        self.data = data
        self.is_async_complete = is_async_complete

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> FunctionTaskParameters:
        function: dict[str, Any] = resource["function"]

        return cls(
            external_id=function["externalId"],
            data=function.get("data"),
            # TODO: Fix: API uses isAsyncComplete and asyncComplete inconsistently:
            is_async_complete=resource.get("isAsyncComplete", resource.get("asyncComplete")),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        function: dict[str, Any] = {
            ("externalId" if camel_case else "external_id"): self.external_id,
        }
        if self.data:
            function["data"] = self.data

        output: dict[str, Any] = {
            "function": function,
        }
        if self.is_async_complete is not None:
            output["isAsyncComplete" if camel_case else "is_async_complete"] = self.is_async_complete
        return output


_SIMULATORS_WARNING = FeaturePreviewWarning(
    api_maturity="General Availability", sdk_maturity="alpha", feature_name="Simulators"
)


class SimulationTaskParameters(WorkflowTaskParameters):
    """
    The simulation parameters are used to specify the simulation routine to be executed.

    Args:
        routine_external_id (str): The external ID of the simulation routine to be executed.
        run_time (int | None): Reference timestamp used for data pre-processing and data sampling.
        inputs (list[SimulationInputOverride] | None): List of input overrides.
    """

    task_type = "simulation"

    def __init__(
        self,
        routine_external_id: str,
        run_time: int | None = None,
        inputs: list[SimulationInputOverride] | None = None,
    ) -> None:
        self.routine_external_id = routine_external_id
        self.run_time = run_time
        self.inputs = inputs

    def __post_init__(self) -> None:
        _SIMULATORS_WARNING.warn()

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> SimulationTaskParameters:
        simulation: dict[str, Any] = resource["simulation"]

        return cls(
            routine_external_id=simulation["routineExternalId"],
            run_time=simulation.get("runTime"),
            inputs=[SimulationInputOverride._load(item) for item in simulation.get("inputs", [])]
            if simulation.get("inputs")
            else None,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        simulation: dict[str, Any] = {
            "routineExternalId" if camel_case else "routine_external_id": self.routine_external_id,
        }
        if self.run_time:
            simulation["runTime" if camel_case else "run_time"] = self.run_time
        if self.inputs:
            simulation["inputs"] = [item.dump(camel_case) for item in self.inputs]

        return {"simulation": simulation}


class TransformationTaskParameters(WorkflowTaskParameters):
    """
    The transformation parameters are used to specify the transformation to be called.

    Args:
        external_id (str): The external ID of the transformation to be called.
        concurrency_policy (Literal['fail', 'restartAfterCurrent', 'waitForCurrent']): Determines the behavior of the task if the Transformation is already running. ``fail``: The task fails if another instance of the Transformation is currently running. ``waitForCurrent``: The task will pause and wait for the already running Transformation to complete. Once completed, the task is completed. This mode is useful for preventing redundant Transformation runs. ``restartAfterCurrent``: The task waits for the ongoing Transformation to finish. After completion, the task restarts the Transformation. This mode ensures that the most recent data can be used by following tasks.
        use_transformation_credentials (bool): If set to `true`, the transformation will be run using the client credentials configured on the transformation. If set to `false`, the transformation will be run using the client credentials used to trigger the workflow.
    """

    task_type = "transformation"

    def __init__(
        self,
        external_id: str,
        concurrency_policy: Literal["fail", "restartAfterCurrent", "waitForCurrent"] = "fail",
        use_transformation_credentials: bool = False,
    ) -> None:
        self.external_id = external_id
        self.concurrency_policy = concurrency_policy
        self.use_transformation_credentials = use_transformation_credentials

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> TransformationTaskParameters:
        return cls(
            resource[cls.task_type]["externalId"],
            resource[cls.task_type].get("concurrencyPolicy", "fail"),
            resource[cls.task_type].get("useTransformationCredentials", False),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        transformation = {
            "externalId" if camel_case else "external_id": self.external_id,
            "concurrencyPolicy" if camel_case else "concurrency_policy": self.concurrency_policy,
            "useTransformationCredentials"
            if camel_case
            else "use_transformation_credentials": self.use_transformation_credentials,
        }

        return {self.task_type: transformation}


class CDFTaskParameters(WorkflowTaskParameters):
    """
    The CDF request parameters are used to specify a request to the Cognite Data Fusion API.

    Args:
        resource_path (str): The resource path of the request. Note the path of the request which is prefixed by '{cluster}.cognitedata.com/api/v1/project/{project}' based on the cluster and project of the request.
        method (Literal['GET', 'POST', 'PUT', 'DELETE'] | str): The HTTP method of the request.
        query_parameters (dict | str | None): The query parameters of the request. Defaults to None.
        body (dict | str | None): The body of the request. Defaults to None. Limited to 1024KiB in size
        request_timeout_in_millis (int | str): The timeout of the request in milliseconds. Defaults to 10000.

    Examples:

        Call the asset/list endpoint with a limit of 10:

            >>> from cognite.client.data_classes import WorkflowTask, CDFTaskParameters
            >>> task = WorkflowTask(
            ...     external_id="task1",
            ...     parameters=CDFTaskParameters(
            ...         resource_path="/assets/list",
            ...         method="GET",
            ...         query_parameters={"limit": 10},
            ...     ),
            ... )

    """

    task_type = "cdf"

    def __init__(
        self,
        resource_path: str,
        method: Literal["GET", "POST", "PUT", "DELETE"] | str,
        query_parameters: dict | str | None = None,
        body: dict | str | None = None,
        request_timeout_in_millis: int | str = 10000,
    ) -> None:
        self.resource_path = resource_path
        self.method = method
        self.query_parameters = query_parameters or {}
        self.body = body or {}
        self.request_timeout_in_millis = request_timeout_in_millis

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> Self:
        cdf_request: dict[str, Any] = resource["cdfRequest"]

        return cls(
            cdf_request["resourcePath"],
            cdf_request["method"],
            cdf_request.get("queryParameters"),
            cdf_request.get("body"),
            cdf_request.get("requestTimeoutInMillis", 10000),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        return {
            ("cdfRequest" if camel_case else "cdf_request"): output,
        }


class SubworkflowTaskParameters(WorkflowTaskParameters):
    """
    The subworkflow task parameters are used to specify a subworkflow task.

    When a workflow is made of stages with dependencies between them, we can use subworkflow tasks for convenience. It takes the tasks parameter which is an array of
    function, transformation, cdf, ..., task definitions. This array needs to be statically set on the workflow definition (if it needs to be defined at runtime, use a
    dynamic task).

    Args:
        tasks (list[WorkflowTask]): The tasks belonging to the subworkflow.
    """

    task_type = "subworkflow"

    def __init__(self, tasks: list[WorkflowTask]) -> None:
        self.tasks = tasks

    @classmethod
    def _load(cls: type[Self], resource: dict, cognite_client: AsyncCogniteClient | None = None) -> Self:
        subworkflow: dict[str, Any] = resource[cls.task_type]

        return cls(
            [WorkflowTask._load(task) for task in subworkflow["tasks"]],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {self.task_type: {"tasks": [task.dump(camel_case) for task in self.tasks]}}


class SubworkflowReferenceParameters(WorkflowTaskParameters):
    """
    The subworkflow task parameters are used to specify a subworkflow task.
    When a workflow is made of stages with dependencies between them, we can use subworkflow tasks for convenience.
    The subworkflow reference is used to specifying a reference to another workflow which will be embedded into the execution at start time.

    Args:
        workflow_external_id (str): The external ID of the referenced workflow.
        version (str): The version of the referenced workflow.
    """

    task_type = "subworkflow"

    def __init__(self, workflow_external_id: str, version: str) -> None:
        self.workflow_external_id = workflow_external_id
        self.version = version

    @classmethod
    def _load(cls: type[Self], resource: dict, cognite_client: AsyncCogniteClient | None = None) -> Self:
        subworkflow: dict[str, Any] = resource[cls.task_type]

        return cls(workflow_external_id=subworkflow["workflowExternalId"], version=subworkflow["version"])

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            self.task_type: {
                "workflowExternalId": self.workflow_external_id,
                "version": self.version,
            }
        }


class DynamicTaskParameters(WorkflowTaskParameters):
    """
    The dynamic task parameters are used to specify a dynamic task.

    When the tasks and their order of execution are determined at runtime, we use dynamic tasks. It takes the tasks parameter which is a Reference to
    an array of function, transformation, and cdf task definitions. This array should be generated and returned by a previous step in the workflow, for instance,
    a Cognite Function task.

    Tip:
        You can reference data from other tasks or the workflow. You do this by following the format
        `${prefix.jsonPath}` in the expression. Some valid option are:

        - `${workflow.input}`: The workflow input.
        - `${<taskExternalId>.output}`: The output of the task with the given external id.
        - `${<taskExternalId>.input}`: The input of the task with the given external id.
        - `${<taskExternalId>.input.someKey}`: A specific key within the input of the task with the given external id.

    Args:
        tasks (list[WorkflowTask] | str): The tasks to be dynamically executed. The dynamic task is a string that is evaluated
                    during the workflow's execution. When calling Version Upsert, the tasks parameter must be a Reference string.
                    When calling Execution details, the tasks parameter will be a list of WorkflowTask objects.
    """

    task_type = "dynamic"

    def __init__(self, tasks: list[WorkflowTask] | str) -> None:
        self.tasks = tasks

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> Self:
        dynamic: dict[str, Any] = resource[cls.task_type]

        # can either be a reference string (i.e., in case of WorkflowDefinitions)
        if isinstance(dynamic["tasks"], str):
            return cls(dynamic["tasks"])

        # or can be resolved to a list of Tasks (i.e., during or after execution)
        return cls(
            [WorkflowTask._load(task) for task in dynamic["tasks"]],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            self.task_type: {
                "tasks": self.tasks if isinstance(self.tasks, str) else [task.dump(camel_case) for task in self.tasks]
            }
        }


class WorkflowTask(CogniteResource):
    """
    This class represents a workflow task.

    Note:
        Tasks do not distinguish between write and read versions.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        parameters (WorkflowTaskParameters): The parameters of the task.
        name (str | None): The name of the task. Defaults to None.
        description (str | None): The description of the task. Defaults to None.
        retries (int): The number of retries for the task. Defaults to 3.
        timeout (int): The timeout of the task in seconds. Defaults to 3600.
        on_failure (Literal['abortWorkflow', 'skipTask']): The policy to handle failures and timeouts. Defaults to *abortWorkflow*. ``skipTask``: For both failures and timeouts, the task will retry until the retries are exhausted. After that, the Task is marked as COMPLETED_WITH_ERRORS and the subsequent tasks are executed. ``abortWorkflow``: In case of failures, retries will be performed until exhausted. After which the task is marked as FAILED and the Workflow is marked the same. In the event of a timeout, no retries are undertaken; the task is marked as TIMED_OUT and the Workflow is marked as FAILED.
        depends_on (list[str] | None): The external ids of the tasks that this task depends on. Defaults to None.
    """

    def __init__(
        self,
        external_id: str,
        parameters: WorkflowTaskParameters,
        name: str | None = None,
        description: str | None = None,
        retries: int = 3,
        timeout: int = 3600,
        on_failure: Literal["abortWorkflow", "skipTask"] = "abortWorkflow",
        depends_on: list[str] | None = None,
    ) -> None:
        self.external_id = external_id
        self.parameters = parameters
        self.name = name
        self.description = description
        self.retries = retries
        self.timeout = timeout
        self.on_failure = on_failure
        self.depends_on = depends_on

    @property
    def type(self) -> ValidTaskType:
        return self.parameters.task_type

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> WorkflowTask:
        return cls(
            external_id=resource["externalId"],
            parameters=WorkflowTaskParameters.load_parameters(resource),
            name=resource.get("name"),
            description=resource.get("description"),
            # Allow default to come from the API.
            retries=resource.get("retries"),  # type: ignore[arg-type]
            timeout=resource.get("timeout"),  # type: ignore[arg-type]
            on_failure=resource.get("onFailure"),  # type: ignore[arg-type]
            depends_on=[dep["externalId"] for dep in resource.get("dependsOn", [])] or None,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = {
            "externalId" if camel_case else "external_id": self.external_id,
            "type": self.type,
            "parameters": self.parameters.dump(camel_case),
        }
        if self.retries is not None:
            output["retries"] = self.retries
        if self.timeout is not None:
            output["timeout"] = self.timeout
        if self.on_failure is not None:
            output["onFailure" if camel_case else "on_failure"] = self.on_failure
        if self.name:
            output["name"] = self.name
        if self.description:
            output["description"] = self.description
        if self.depends_on:
            output[("dependsOn" if camel_case else "depends_on")] = [
                {("externalId" if camel_case else "external_id"): dependency} for dependency in self.depends_on
            ]
        return output


class WorkflowTaskOutput(ABC):
    task_type: ClassVar[str]

    @classmethod
    @abstractmethod
    def load(cls, data: dict) -> Self:
        raise NotImplementedError

    @classmethod
    def load_output(cls, data: dict) -> WorkflowTaskOutput:
        task_type = data["taskType"]
        if task_type == "function":
            return FunctionTaskOutput.load(data)
        elif task_type == "transformation":
            return TransformationTaskOutput.load(data)
        elif task_type == "cdf":
            return CDFTaskOutput.load(data)
        elif task_type == "dynamic":
            return DynamicTaskOutput.load(data)
        elif task_type == "subworkflow":
            return SubworkflowTaskOutput.load(data)
        elif task_type == "simulation":
            return SimulationTaskOutput.load(data)
        else:
            raise ValueError(f"Unknown task type: {task_type}")

    @abstractmethod
    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        raise NotImplementedError


class FunctionTaskOutput(WorkflowTaskOutput):
    """
    The class represent the output of Cognite Function task.

    Args:
        call_id (int | None): The call_id of the CDF Function call.
        function_id (int | None): The function_id of the CDF Function.
        response (dict | None): The response of the CDF Function call.

    """

    task_type: ClassVar[str] = "function"

    def __init__(self, call_id: int | None, function_id: int | None, response: dict | None) -> None:
        self.call_id = call_id
        self.function_id = function_id
        self.response = response

    @classmethod
    def load(cls, data: dict[str, Any]) -> FunctionTaskOutput:
        output = data["output"]
        return cls(output.get("callId"), output.get("functionId"), output.get("response"))

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "callId" if camel_case else "call_id": self.call_id,
            "functionId" if camel_case else "function_id": self.function_id,
            "response": self.response,
        }


class SimulationTaskOutput(WorkflowTaskOutput):
    """
    The class represent the output of Simulation execution.

    Args:
        run_id (int | None): The run ID of the simulation run.
        log_id (int | None): The log ID of the simulation run.
        status_message (str | None): Status message of the simulation execution.
    """

    task_type: ClassVar[str] = "simulation"

    def __post_init__(self) -> None:
        _SIMULATORS_WARNING.warn()

    def __init__(
        self,
        run_id: int | None,
        log_id: int | None,
        status_message: str | None,
    ) -> None:
        self.run_id = run_id
        self.log_id = log_id
        self.status_message = status_message

    @classmethod
    def load(cls, data: dict[str, Any]) -> SimulationTaskOutput:
        output = data["output"]
        return cls(
            run_id=output.get("runId"),
            log_id=output.get("logId"),
            status_message=output.get("statusMessage"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "runId" if camel_case else "run_id": self.run_id,
            "logId" if camel_case else "log_id": self.log_id,
            "statusMessage" if camel_case else "status_message": self.status_message,
        }


class TransformationTaskOutput(WorkflowTaskOutput):
    """
    The transformation output is used to specify the output of a transformation task.

    Args:
        job_id (int | None): The job id of the transformation job.
    """

    task_type: ClassVar[str] = "transformation"

    def __init__(self, job_id: int | None) -> None:
        self.job_id = job_id

    @classmethod
    def load(cls, data: dict[str, Any]) -> TransformationTaskOutput:
        output = data["output"]
        return cls(output.get("jobId"))

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {("jobId" if camel_case else "job_id"): self.job_id}


class CDFTaskOutput(WorkflowTaskOutput):
    """
    The CDF Request output is used to specify the output of a CDF Request.

    Args:
        response (str | dict | None): The response of the CDF Request. Will be a JSON object if content-type is application/json, otherwise will be a string.
        status_code (int | None): The status code of the CDF Request.
    """

    task_type: ClassVar[str] = "cdf"

    def __init__(self, response: str | dict | None, status_code: int | None) -> None:
        self.response = response
        self.status_code = status_code

    @classmethod
    def load(cls, data: dict[str, Any]) -> CDFTaskOutput:
        output = data["output"]
        return cls(output.get("response"), output.get("statusCode"))

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "response": self.response,
            ("statusCode" if camel_case else "status_code"): self.status_code,
        }


class DynamicTaskOutput(WorkflowTaskOutput):
    """
    The dynamic task output is used to specify the output of a dynamic task.
    """

    task_type: ClassVar[str] = "dynamic"

    def __init__(self) -> None: ...

    @classmethod
    def load(cls, data: dict[str, Any]) -> DynamicTaskOutput:
        return cls()

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {}


class SubworkflowTaskOutput(WorkflowTaskOutput):
    """
    The subworkflow task output is used to specify the output of a subworkflow task.
    """

    task_type: ClassVar[str] = "subworkflow"

    def __init__(self) -> None: ...

    @classmethod
    def load(cls, data: dict[str, Any]) -> SubworkflowTaskOutput:
        return cls()

    def dump(self, camel_case: bool = False) -> dict[str, Any]:
        return {}


class WorkflowTaskExecution(CogniteObject):
    """
    This class represents a task execution.

    Args:
        id (str): The server generated id of the task execution.
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        status (TaskStatus): The status of the task execution.
        input (WorkflowTaskParameters): The input parameters of the task execution.
        output (WorkflowTaskOutput): The output of the task execution.
        version (str | None): The version of the task execution. Defaults to None.
        start_time (int | None): The start time of the task execution. Unix timestamp in milliseconds. Defaults to None.
        end_time (int | None): The end time of the task execution. Unix timestamp in milliseconds. Defaults to None.
        reason_for_incompletion (str | None): Provides the reason if the workflow did not complete successfully. Defaults to None.
    """

    def __init__(
        self,
        id: str,
        external_id: str,
        status: TaskStatus,
        input: WorkflowTaskParameters,
        output: WorkflowTaskOutput,
        version: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        reason_for_incompletion: str | None = None,
    ) -> None:
        self.id = id
        self.external_id = external_id
        self.status = status
        self.input = input
        self.output = output
        self.version = version
        self.start_time = start_time
        self.end_time = end_time
        self.reason_for_incompletion = reason_for_incompletion

    @property
    def task_type(self) -> ValidTaskType:
        return self.input.task_type

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> WorkflowTaskExecution:
        return cls(
            id=resource["id"],
            external_id=resource["externalId"],
            status=cast(TaskStatus, to_snake_case(resource["status"])),
            input=WorkflowTaskParameters.load_parameters(resource),
            output=WorkflowTaskOutput.load_output(resource),
            version=resource.get("version"),
            start_time=resource.get("startTime"),
            end_time=resource.get("endTime"),
            reason_for_incompletion=resource.get("reasonForIncompletion"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = super().dump(camel_case)
        output["input"] = self.input.dump(camel_case)
        output["status"] = self.status.upper()
        output[("taskType" if camel_case else "task_type")] = self.task_type
        # API uses isAsyncComplete and asyncComplete inconsistently:
        if self.task_type == "function":
            if (is_async_complete := output["input"].get("isAsyncComplete")) is not None:
                output["input"]["asyncComplete"] = is_async_complete
                del output["input"]["isAsyncComplete"]

        output["output"] = self.output.dump(camel_case)
        return output


class WorkflowDefinitionCore(WriteableCogniteResource["WorkflowDefinitionUpsert"], ABC):
    """This class represents a workflow definition.

    A workflow definition defines the tasks and order/dependencies of these tasks.

    Args:
        tasks (list[WorkflowTask]): The tasks of the workflow definition.
        description (str | None): The description of the workflow definition. Note that when updating a workflow definition
                            description, it will always be overwritten also if it is set to None. Meaning if the
                            workflow definition already has a description, and you want to keep it, you need to provide
                            the description when updating it.
    """

    def __init__(
        self,
        tasks: list[WorkflowTask],
        description: str | None = None,
    ) -> None:
        self.tasks = tasks
        self.description = description

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            tasks=[WorkflowTask._load(task) for task in resource["tasks"]],
            description=resource.get("description"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = {"tasks": [task.dump(camel_case) for task in self.tasks]}
        if self.description:
            output["description"] = self.description
        return output


class WorkflowDefinitionUpsert(WorkflowDefinitionCore):
    """
    This class represents a workflow definition. This represents the write/update version of a workflow definition.

    A workflow definition defines the tasks and order/dependencies of these tasks.

    Args:
        tasks (list[WorkflowTask]): The tasks of the workflow definition.
        description (str | None): The description of the workflow definition. Note that when updating a workflow definition
                            description, it will always be overwritten also if it is set to None. Meaning if the
                            workflow definition already has a description, and you want to keep it, you need to provide
                            the description when updating it.
    """

    def __init__(self, tasks: list[WorkflowTask], description: str | None = None) -> None:
        super().__init__(tasks, description)

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> WorkflowDefinitionUpsert:
        return cls(
            tasks=[WorkflowTask._load(task) for task in resource["tasks"]],
            description=resource.get("description"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = {"tasks": [task.dump(camel_case) for task in self.tasks]}
        if self.description:
            output["description"] = self.description
        return output

    def as_write(self) -> WorkflowDefinitionUpsert:
        """Returns this WorkflowDefinitionUpsert in its write format."""
        return self


class WorkflowDefinition(WorkflowDefinitionCore):
    """
    This class represents a workflow definition. This represents the read version of a workflow definition.

    A workflow definition defines the tasks and order/dependencies of these tasks.

    Args:
        hash_ (str): The hash of the tasks and description. This is used to uniquely identify the workflow definition as you can overwrite a workflow version.
        tasks (list[WorkflowTask]): The tasks of the workflow definition.
        description (str | None): The description of the workflow definition. Defaults to None.
    """

    def __init__(
        self,
        hash_: str,
        tasks: list[WorkflowTask],
        description: str | None = None,
    ) -> None:
        super().__init__(tasks, description)
        self.hash_ = hash_

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> WorkflowDefinition:
        return cls(
            hash_=resource["hash"],
            tasks=[WorkflowTask._load(task) for task in resource["tasks"]],
            description=resource.get("description"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        output["hash"] = self.hash_
        return output

    def as_write(self) -> WorkflowDefinitionUpsert:
        """Returns this WorkflowDefinition in its write format."""
        return WorkflowDefinitionUpsert(
            tasks=self.tasks,
            description=self.description,
        )


class WorkflowVersionCore(WriteableCogniteResource["WorkflowVersionUpsert"], ABC):
    """
    This class represents a workflow version.

    Args:
        workflow_external_id (str): The external ID of the workflow.
        version (str): The version of the workflow.
    """

    def __init__(
        self,
        workflow_external_id: str,
        version: str,
    ) -> None:
        self.workflow_external_id = workflow_external_id
        self.version = version

    def as_id(self) -> WorkflowVersionId:
        return WorkflowVersionId(
            workflow_external_id=self.workflow_external_id,
            version=self.version,
        )


class WorkflowVersionUpsert(WorkflowVersionCore):
    """
    This class represents a workflow version. This is the write-variant, used when creating or updating a workflow variant.

    Args:
        workflow_external_id (str): The external ID of the workflow.
        version (str): The version of the workflow.
        workflow_definition (WorkflowDefinitionUpsert): The workflow definition of the workflow version.

    """

    def __init__(self, workflow_external_id: str, version: str, workflow_definition: WorkflowDefinitionUpsert) -> None:
        super().__init__(
            workflow_external_id=workflow_external_id,
            version=version,
        )
        self.workflow_definition = workflow_definition

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> Self:
        workflow_definition: dict[str, Any] = resource["workflowDefinition"]
        return cls(
            workflow_external_id=resource["workflowExternalId"],
            version=resource["version"],
            workflow_definition=WorkflowDefinitionUpsert._load(workflow_definition),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            ("workflowExternalId" if camel_case else "workflow_external_id"): self.workflow_external_id,
            "version": self.version,
            ("workflowDefinition" if camel_case else "workflow_definition"): self.workflow_definition.dump(camel_case),
        }

    def as_write(self) -> WorkflowVersionUpsert:
        """Returns this WorkflowVersionUpsert instance."""
        return self


class WorkflowVersion(WorkflowVersionCore):
    """
    This class represents a workflow version. This is the read variant, used when retrieving/listing a workflow variant.

    Args:
        workflow_external_id (str): The external ID of the workflow.
        version (str): The version of the workflow.
        workflow_definition (WorkflowDefinition): The workflow definition of the workflow version.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
    """

    def __init__(
        self,
        workflow_external_id: str,
        version: str,
        workflow_definition: WorkflowDefinition,
        created_time: int,
        last_updated_time: int,
    ) -> None:
        super().__init__(workflow_external_id=workflow_external_id, version=version)
        self.workflow_definition = workflow_definition
        self.created_time = created_time
        self.last_updated_time = last_updated_time

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> WorkflowVersion:
        workflow_definition: dict[str, Any] = resource["workflowDefinition"]
        return cls(
            workflow_external_id=resource["workflowExternalId"],
            version=resource["version"],
            workflow_definition=WorkflowDefinition._load(workflow_definition),
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "workflowExternalId" if camel_case else "workflow_external_id": self.workflow_external_id,
            "version": self.version,
            "workflowDefinition" if camel_case else "workflow_definition": self.workflow_definition.dump(camel_case),
            "createdTime" if camel_case else "created_time": self.created_time,
            "lastUpdatedTime" if camel_case else "last_updated_time": self.last_updated_time,
        }

    def as_write(self) -> WorkflowVersionUpsert:
        """Returns a WorkflowVersionUpsert object with the same data."""
        return WorkflowVersionUpsert(
            workflow_external_id=self.workflow_external_id,
            version=self.version,
            workflow_definition=self.workflow_definition.as_write(),
        )


class WorkflowVersionUpsertList(CogniteResourceList[WorkflowVersionUpsert]):
    """This class represents a list of workflow versions."""

    _RESOURCE = WorkflowVersionUpsert

    def as_ids(self) -> WorkflowIds:
        """Returns a WorkflowIds object with the workflow version ids."""
        return WorkflowIds([workflow_version.as_id() for workflow_version in self.data])


class WorkflowVersionList(WriteableCogniteResourceList[WorkflowVersionUpsert, WorkflowVersion]):
    """
    This class represents a list of workflow versions.
    """

    _RESOURCE = WorkflowVersion

    def as_ids(self) -> WorkflowIds:
        """Returns a WorkflowIds object with the workflow version ids."""
        return WorkflowIds([workflow_version.as_id() for workflow_version in self.data])

    def as_write(self) -> WorkflowVersionUpsertList:
        """Returns a WorkflowVersionUpsertList object with the same data."""
        return WorkflowVersionUpsertList([workflow_version.as_write() for workflow_version in self.data])


class WorkflowExecution(CogniteResource):
    """
    This class represents a workflow execution.

    Args:
        id (str): The server generated id of the workflow execution.
        workflow_external_id (str): The external ID of the workflow.
        status (WorkflowStatus): The status of the workflow execution.
        created_time (int): The time when the workflow execution was created. Unix timestamp in milliseconds.
        version (str | None): The version of the workflow. Defaults to None.
        start_time (int | None): The start time of the workflow execution. Unix timestamp in milliseconds. Defaults to None.
        end_time (int | None): The end time of the workflow execution. Unix timestamp in milliseconds. Defaults to None.
        reason_for_incompletion (str | None): Provides the reason if the workflow did not complete successfully. Defaults to None.
        metadata (dict | None): Application specific metadata.
    """

    def __init__(
        self,
        id: str,
        workflow_external_id: str,
        status: WorkflowStatus,
        created_time: int,
        version: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        reason_for_incompletion: str | None = None,
        metadata: dict | None = None,
    ) -> None:
        self.id = id
        self.workflow_external_id = workflow_external_id
        self.version = version
        self.status = status
        self.created_time = created_time
        self.start_time = start_time
        self.end_time = end_time
        self.reason_for_incompletion = reason_for_incompletion
        self.metadata = metadata

    def as_workflow_id(self) -> WorkflowVersionId:
        return WorkflowVersionId(workflow_external_id=self.workflow_external_id, version=self.version)

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> WorkflowExecution:
        return cls(
            id=resource["id"],
            workflow_external_id=resource["workflowExternalId"],
            version=resource.get("version"),
            status=cast(
                WorkflowStatus,
                to_snake_case(resource["status"]),
            ),
            created_time=resource["createdTime"],
            start_time=resource.get("startTime"),
            end_time=resource.get("endTime"),
            reason_for_incompletion=resource.get("reasonForIncompletion"),
            metadata=resource.get("metadata"),
        )


class WorkflowExecutionList(CogniteResourceList[WorkflowExecution], InternalIdTransformerMixin):
    """
    This class represents a list of workflow executions.
    """

    _RESOURCE = WorkflowExecution


class WorkflowExecutionDetailed(WorkflowExecution):
    """
    This class represents a detailed workflow execution.

    A detailed workflow execution contains the input and output of each task in the workflow execution. In addition,
    it contains the workflow definition of the workflow.

    Args:
        id (str): The server generated id of the workflow execution.
        workflow_external_id (str): The external ID of the workflow.
        workflow_definition (WorkflowDefinition): The workflow definition of the workflow.
        status (WorkflowStatus): The status of the workflow execution.
        executed_tasks (list[WorkflowTaskExecution]): The executed tasks of the workflow execution.
        created_time (int): The time when the workflow execution was created. Unix timestamp in milliseconds.
        version (str | None): The version of the workflow. Defaults to None.
        start_time (int | None): The start time of the workflow execution. Unix timestamp in milliseconds. Defaults to None.
        end_time (int | None): The end time of the workflow execution. Unix timestamp in milliseconds. Defaults to None.
        reason_for_incompletion (str | None): Provides the reason if the workflow did not complete successfully. Defaults to None.
        input (dict | None): Input arguments the workflow was triggered with.
        metadata (dict | None): Metadata set when the workflow was triggered.
    """

    def __init__(
        self,
        id: str,
        workflow_external_id: str,
        workflow_definition: WorkflowDefinition,
        status: WorkflowStatus,
        executed_tasks: list[WorkflowTaskExecution],
        created_time: int,
        version: str | None = None,
        start_time: int | None = None,
        end_time: int | None = None,
        reason_for_incompletion: str | None = None,
        input: dict | None = None,
        metadata: dict | None = None,
    ) -> None:
        super().__init__(
            id, workflow_external_id, status, created_time, version, start_time, end_time, reason_for_incompletion
        )
        self.workflow_definition = workflow_definition
        self.executed_tasks = executed_tasks
        self.input = input
        self.metadata = metadata

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> WorkflowExecutionDetailed:
        return cls(
            id=resource["id"],
            workflow_external_id=resource["workflowExternalId"],
            version=resource.get("version"),
            status=cast(
                WorkflowStatus,
                to_snake_case(resource["status"]),
            ),
            created_time=resource["createdTime"],
            start_time=resource.get("startTime"),
            end_time=resource.get("endTime"),
            reason_for_incompletion=resource.get("reasonForIncompletion"),
            workflow_definition=WorkflowDefinition._load(resource["workflowDefinition"]),
            executed_tasks=[WorkflowTaskExecution.load(task) for task in resource["executedTasks"]],
            input=resource.get("input"),
            metadata=resource.get("metadata"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        output[("workflowDefinition" if camel_case else "workflow_definition")] = self.workflow_definition.dump(
            camel_case
        )
        output[("executedTasks" if camel_case else "executed_tasks")] = [
            task.dump(camel_case) for task in self.executed_tasks
        ]
        if self.input:
            output["input"] = self.input
        if self.metadata:
            output["metadata"] = self.metadata
        return output

    def as_execution(self) -> WorkflowExecution:
        return WorkflowExecution(
            id=self.id,
            workflow_external_id=self.workflow_external_id,
            version=self.version,
            status=self.status,
            created_time=self.created_time,
            start_time=self.start_time,
            end_time=self.end_time,
            reason_for_incompletion=self.reason_for_incompletion,
            metadata=self.metadata,
        )


@dataclass(frozen=True)
class WorkflowVersionId:
    """
    This class represents a Workflow Version Identifier.

    Args:
        workflow_external_id (str): The external ID of the workflow.
        version (str, optional): The version of the workflow. Defaults to None.
    """

    workflow_external_id: str
    version: str | None = None

    def as_tuple(self) -> tuple[str, str | None]:
        return self.workflow_external_id, self.version

    @classmethod
    def load(cls, resource: dict) -> Self:
        if "workflowExternalId" in resource:
            workflow_external_id = resource["workflowExternalId"]
        elif "externalId" in resource:
            workflow_external_id = resource["externalId"]
        else:
            raise ValueError("Invalid input to WorkflowVersionId.load")

        return cls(workflow_external_id=workflow_external_id, version=resource.get("version"))

    def dump(self, camel_case: bool = True, as_external_id_key: bool = False) -> dict[str, Any]:
        if as_external_id_key:
            output: dict[str, Any] = {"externalId" if camel_case else "external_id": self.workflow_external_id}
        else:
            output = {"workflowExternalId" if camel_case else "workflow_external_id": self.workflow_external_id}
        if self.version:
            output["version"] = self.version
        return output


class WorkflowIds(UserList):
    """This class represents a list of Workflow Version Identifiers."""

    def __init__(self, workflow_ids: Collection[WorkflowVersionId]) -> None:
        for workflow_id in workflow_ids:
            if not isinstance(workflow_id, WorkflowVersionId):
                raise TypeError(
                    f"All resources for class '{type(self).__name__}' must be of type "
                    f"'{type(WorkflowVersionId).__name__}', not '{type(workflow_id)}'."
                )
        super().__init__(workflow_ids)

    def as_tuples(self) -> list[tuple[str, str | None]]:
        return [wid.as_tuple() for wid in self]

    @classmethod
    def load(cls, resource: Any) -> Self:
        workflow_ids: Sequence[WorkflowVersionId]
        if isinstance(resource, tuple) and len(resource) == 2 and all(isinstance(x, str) for x in resource):
            workflow_ids = [WorkflowVersionId(*resource)]
        elif isinstance(resource, WorkflowVersionId):
            workflow_ids = [resource]
        elif isinstance(resource, str):
            workflow_ids = [WorkflowVersionId(workflow_external_id=resource)]
        elif isinstance(resource, dict):
            workflow_ids = [WorkflowVersionId.load(resource)]
        elif isinstance(resource, Sequence) and resource:
            workflow_ids = []
            for wf in resource:
                match wf:
                    case tuple():
                        workflow_ids.append(WorkflowVersionId(*wf))
                    case str():
                        workflow_ids.append(WorkflowVersionId(workflow_external_id=wf))
                    case _:
                        workflow_ids.append(wf)
        else:
            raise ValueError("Invalid input to WorkflowIds.load")
        return cls(workflow_ids)

    def dump(self, camel_case: bool = True, as_external_id: bool = False) -> list[dict[str, Any]]:
        return [workflow_id.dump(camel_case, as_external_id_key=as_external_id) for workflow_id in self.data]


class WorkflowTriggerRule(CogniteObject, ABC):
    """This is the base class for all workflow trigger rules."""

    _trigger_type: ClassVar[str]

    @final
    @property
    def trigger_type(self) -> str:
        return self._trigger_type

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        dumped = super().dump(camel_case)
        dumped["triggerType" if camel_case else "trigger_type"] = self.trigger_type
        return dumped

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> Self:
        trigger_type = resource["triggerType"]
        if trigger_type in _TRIGGER_RULE_BY_TYPE:
            return cast(Self, _TRIGGER_RULE_BY_TYPE[trigger_type]._load_trigger(resource))
        # If more triggers are added in the future, this ensures that the SDK does not break.
        return cast(Self, UnknownCogniteObject(resource))

    @classmethod
    @abstractmethod
    def _load_trigger(cls, data: dict) -> Self:
        raise NotImplementedError


class WorkflowTriggerDataModelingQuery(Query):
    r"""This class represents a data modeling trigger query.

    Args:
        with_ (dict[str, ResultSetExpression]): A dictionary of result set expressions to use in the query. The keys are used to reference the result set expressions in the select.
        select (dict[str, Select]): A dictionary of select expressions to use in the query. The keys must match the keys in the with\_ dictionary. The select expressions define which properties to include in the result set.
    """

    def __init__(
        self,
        with_: dict[str, ResultSetExpression],
        select: dict[str, Select],
    ) -> None:
        super().__init__(with_, select)
        # Parameters and cursors are not supported for workflow trigger queries:
        self.parameters = None  # type: ignore [assignment]
        self.cursors = None  # type: ignore [assignment]

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            with_={k: ResultSetExpression.load(v) for k, v in resource["with"].items()},
            select={k: Select.load(v) for k, v in resource["select"].items()},
        )


class WorkflowScheduledTriggerRule(WorkflowTriggerRule):
    """
    This class represents a scheduled trigger rule.

    Args:
        cron_expression (str): The cron specification for the scheduled trigger.
        timezone (ZoneInfo | None): The timezone in which the scheduled trigger should be evaluated.
            If not provided, UTC will be used as the default timezone on the server side.
    """

    _trigger_type = "schedule"

    def __init__(self, cron_expression: str, timezone: ZoneInfo | None = None) -> None:
        self.cron_expression = cron_expression
        self.timezone = timezone

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        # Override dump to handle timezone field specially:
        # 1. Only include timezone key when it has a value (avoid "timezone": null)
        # 2. Convert ZoneInfo object to its string key representation
        item = super().dump(camel_case)
        if self.timezone:
            item["timezone"] = self.timezone.key
        return item

    @classmethod
    def _load_trigger(cls, data: dict) -> WorkflowScheduledTriggerRule:
        # Convert timezone to ZoneInfo
        timezone = ZoneInfo(data["timezone"]) if "timezone" in data else None
        return cls(cron_expression=data["cronExpression"], timezone=timezone)


class WorkflowDataModelingTriggerRule(WorkflowTriggerRule):
    """
    This class represents a data modeling trigger rule.

    Args:
        data_modeling_query (WorkflowTriggerDataModelingQuery): The data modeling query of the trigger.
        batch_size (int | None): The batch size of the trigger.
        batch_timeout (int | None): The batch timeout of the trigger.
    """

    _trigger_type = "dataModeling"

    def __init__(
        self,
        data_modeling_query: WorkflowTriggerDataModelingQuery,
        batch_size: int | None = None,
        batch_timeout: int | None = None,
    ) -> None:
        self.data_modeling_query = data_modeling_query
        self.batch_size = batch_size
        self.batch_timeout = batch_timeout

    @classmethod
    def _load_trigger(cls, data: dict) -> WorkflowDataModelingTriggerRule:
        return cls(
            data_modeling_query=WorkflowTriggerDataModelingQuery.load(data["dataModelingQuery"]),
            batch_size=data.get("batchSize"),
            batch_timeout=data.get("batchTimeout"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        item: dict[str, Any] = {
            "trigger_type": self.trigger_type,
            "data_modeling_query": self.data_modeling_query.dump(camel_case=camel_case),
            "batch_size": self.batch_size,
            "batch_timeout": self.batch_timeout,
        }
        if camel_case:
            return convert_all_keys_to_camel_case(item)
        return item


_TRIGGER_RULE_BY_TYPE: dict[str, type[WorkflowTriggerRule]] = {
    subclass._trigger_type: subclass  # type: ignore
    for subclass in WorkflowTriggerRule.__subclasses__()
}


class WorkflowTriggerCore(WriteableCogniteResource["WorkflowTriggerUpsert"], ABC):
    """
    This class represents a base class for a workflow trigger.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        trigger_rule (WorkflowTriggerRule): The trigger rule of the workflow version trigger.
        workflow_external_id (str): The external ID of the workflow.
        workflow_version (str): The version of the workflow.
    """

    def __init__(
        self, external_id: str, trigger_rule: WorkflowTriggerRule, workflow_external_id: str, workflow_version: str
    ) -> None:
        self.external_id = external_id
        self.trigger_rule = trigger_rule
        self.workflow_external_id = workflow_external_id
        self.workflow_version = workflow_version


class WorkflowTriggerUpsert(WorkflowTriggerCore):
    """
    This class represents a workflow trigger for upsertion.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        trigger_rule (WorkflowTriggerRule): The trigger rule of the workflow version trigger.
        workflow_external_id (str): The external ID of the workflow.
        workflow_version (str): The version of the workflow.
        input (dict | None): The input data of the workflow version trigger. Defaults to None.
        metadata (dict | None): Application specific metadata. Defaults to None.
    """

    def __init__(
        self,
        external_id: str,
        trigger_rule: WorkflowTriggerRule,
        workflow_external_id: str,
        workflow_version: str,
        input: dict | None = None,
        metadata: dict | None = None,
    ) -> None:
        super().__init__(
            external_id=external_id,
            trigger_rule=trigger_rule,
            workflow_external_id=workflow_external_id,
            workflow_version=workflow_version,
        )
        self.input = input
        self.metadata = metadata

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        item: dict[str, Any] = {
            "external_id": self.external_id,
            "trigger_rule": self.trigger_rule.dump(camel_case=camel_case),
            "workflow_external_id": self.workflow_external_id,
            "workflow_version": self.workflow_version,
        }

        if self.input:
            item["input"] = self.input
        if self.metadata:
            item["metadata"] = self.metadata
        if camel_case:
            return convert_all_keys_to_camel_case(item)
        return item

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> WorkflowTriggerUpsert:
        return cls(
            external_id=resource["externalId"],
            workflow_external_id=resource["workflowExternalId"],
            workflow_version=resource["workflowVersion"],
            trigger_rule=WorkflowTriggerRule._load(resource["triggerRule"]),
            input=resource.get("input"),
            metadata=resource.get("metadata"),
        )

    def as_write(self) -> WorkflowTriggerUpsert:
        """Returns this workflow trigger create instance."""
        return self


class WorkflowTrigger(WorkflowTriggerCore):
    """
    This class represents a workflow trigger.

    Args:
        external_id (str): The external ID provided by the client. Must be unique for the resource type.
        trigger_rule (WorkflowTriggerRule): The trigger rule of the workflow version trigger.
        workflow_external_id (str): The external ID of the workflow.
        workflow_version (str): The version of the workflow.
        is_paused (bool): Whether the trigger is paused.
        input (dict | None): The input data passed to the workflow when an execution is started. Defaults to None.
        metadata (dict | None): Application specific metadata. Defaults to None.
        created_time (int): The time when the workflow version trigger was created. Unix timestamp in milliseconds. Defaults to None.
        last_updated_time (int): The time when the workflow version trigger was last updated. Unix timestamp in milliseconds. Defaults to None.
    """

    def __init__(
        self,
        external_id: str,
        trigger_rule: WorkflowTriggerRule,
        workflow_external_id: str,
        workflow_version: str,
        is_paused: bool,
        input: dict | None,
        metadata: dict | None,
        created_time: int,
        last_updated_time: int,
    ) -> None:
        super().__init__(
            external_id=external_id,
            trigger_rule=trigger_rule,
            workflow_external_id=workflow_external_id,
            workflow_version=workflow_version,
        )
        self.input = input or {}
        self.metadata = metadata or {}
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self.is_paused = is_paused

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        item: dict[str, Any] = {
            "external_id": self.external_id,
            "trigger_rule": self.trigger_rule.dump(camel_case=camel_case),
            "workflow_external_id": self.workflow_external_id,
            "workflow_version": self.workflow_version,
        }
        if self.input:
            item["input"] = self.input
        if self.metadata:
            item["metadata"] = self.metadata
        if self.created_time:
            item["created_time"] = self.created_time
        if self.last_updated_time:
            item["last_updated_time"] = self.last_updated_time
        item["is_paused"] = self.is_paused
        if camel_case:
            return convert_all_keys_to_camel_case(item)
        return item

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> WorkflowTrigger:
        return cls(
            external_id=resource["externalId"],
            workflow_external_id=resource["workflowExternalId"],
            workflow_version=resource["workflowVersion"],
            trigger_rule=WorkflowTriggerRule._load(resource["triggerRule"]),
            input=resource.get("input"),
            metadata=resource.get("metadata"),
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
            is_paused=resource["isPaused"],
        )

    def as_write(self) -> WorkflowTriggerUpsert:
        """Returns this workflow trigger instance."""
        return WorkflowTriggerUpsert(
            external_id=self.external_id,
            trigger_rule=self.trigger_rule,
            workflow_external_id=self.workflow_external_id,
            workflow_version=self.workflow_version,
            input=self.input,
            metadata=self.metadata,
        )


class WorkflowTriggerUpsertList(CogniteResourceList[WorkflowTriggerUpsert], ExternalIDTransformerMixin):
    _RESOURCE = WorkflowTriggerUpsert


class WorkflowTriggerList(
    WriteableCogniteResourceList[WorkflowTriggerUpsert, WorkflowTrigger], ExternalIDTransformerMixin
):
    """
    This class represents a list of workflow triggers.
    """

    _RESOURCE = WorkflowTrigger

    def as_write(self) -> WorkflowTriggerUpsertList:
        """Returns a WorkflowTriggerUpsertList object with the same data."""
        return WorkflowTriggerUpsertList([workflow_trigger.as_write() for workflow_trigger in self.data])


class WorkflowTriggerRun(CogniteResource):
    """
    This class represents a workflow trigger run.
    """

    def __init__(
        self,
        external_id: str,
        fire_time: int,
        workflow_external_id: str,
        workflow_version: str,
        status: Literal["success", "failed"],
        workflow_execution_id: str | None,
        reason_for_failure: str | None,
    ) -> None:
        self.external_id = external_id
        self.fire_time = fire_time
        self.workflow_external_id = workflow_external_id
        self.workflow_version = workflow_version
        self.workflow_execution_id = workflow_execution_id
        self.status = status
        self.reason_for_failure = reason_for_failure

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        item = {
            "external_id": self.external_id,
            "fire_time": self.fire_time,
            "workflow_external_id": self.workflow_external_id,
            "workflow_version": self.workflow_version,
            "status": self.status,
        }
        if self.workflow_execution_id:
            item["workflow_execution_id"] = self.workflow_execution_id
        if self.reason_for_failure:
            item["reason_for_failure"] = self.reason_for_failure
        if camel_case:
            return convert_all_keys_to_camel_case(item)
        return item

    @classmethod
    def _load(cls, resource: dict, cognite_client: AsyncCogniteClient | None = None) -> WorkflowTriggerRun:
        return cls(
            external_id=resource["externalId"],
            fire_time=resource["fireTime"],
            workflow_external_id=resource["workflowExternalId"],
            workflow_version=resource["workflowVersion"],
            status=resource["status"],
            workflow_execution_id=resource.get("workflowExecutionId"),
            reason_for_failure=resource.get("reasonForFailure"),
        )


class WorkflowTriggerRunList(CogniteResourceList[WorkflowTriggerRun], ExternalIDTransformerMixin):
    """This class represents a list of workflow trigger runs."""

    _RESOURCE = WorkflowTriggerRun



"""
===============================================================================
ffb8afc8b8e4dd416046292bf4dddc81
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import overload

from cognite.client import AsyncCogniteClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes import (
    TransformationNotification,
    TransformationNotificationList,
    TransformationNotificationWrite,
)
from cognite.client.utils._async_helpers import SyncIterator, run_sync


class SyncTransformationNotificationsAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client

    @overload
    def __call__(
        self,
        chunk_size: None = None,
        transformation_id: int | None = None,
        transformation_external_id: str | None = None,
        destination: str | None = None,
        limit: int | None = None,
    ) -> Iterator[TransformationNotification]: ...

    @overload
    def __call__(
        self,
        chunk_size: int,
        transformation_id: int | None = None,
        transformation_external_id: str | None = None,
        destination: str | None = None,
        limit: int | None = None,
    ) -> Iterator[TransformationNotificationList]: ...

    def __call__(
        self,
        chunk_size: int | None = None,
        transformation_id: int | None = None,
        transformation_external_id: str | None = None,
        destination: str | None = None,
        limit: int | None = None,
    ) -> Iterator[TransformationNotification] | Iterator[TransformationNotificationList]:
        """
        Iterate over transformation notifications

        Args:
            chunk_size (int | None): Number of notifications to yield per chunk. Defaults to yielding notifications one by one.
            transformation_id (int | None): Filter by transformation internal numeric ID.
            transformation_external_id (str | None): Filter by transformation externalId.
            destination (str | None): Filter by notification destination.
            limit (int | None): Limits the number of results to be returned. Defaults to yielding all notifications.

        Yields:
            TransformationNotification | TransformationNotificationList: Yields notifications one by one if chunk_size is None, otherwise yields lists of notifications.
        """  # noqa: DOC404
        yield from SyncIterator(
            self.__async_client.transformations.notifications(
                chunk_size=chunk_size,
                transformation_id=transformation_id,
                transformation_external_id=transformation_external_id,
                destination=destination,
                limit=limit,
            )
        )  # type: ignore [misc]

    @overload
    def create(
        self, notification: TransformationNotification | TransformationNotificationWrite
    ) -> TransformationNotification: ...

    @overload
    def create(
        self, notification: Sequence[TransformationNotification] | Sequence[TransformationNotificationWrite]
    ) -> TransformationNotificationList: ...

    def create(
        self,
        notification: TransformationNotification
        | TransformationNotificationWrite
        | Sequence[TransformationNotification]
        | Sequence[TransformationNotificationWrite],
    ) -> TransformationNotification | TransformationNotificationList:
        """
        `Subscribe for notifications on the transformation errors. <https://developer.cognite.com/api#tag/Transformation-Notifications/operation/createTransformationNotifications>`_

        Args:
            notification (TransformationNotification | TransformationNotificationWrite | Sequence[TransformationNotification] | Sequence[TransformationNotificationWrite]): Notification or list of notifications to create.

        Returns:
            TransformationNotification | TransformationNotificationList: Created notification(s)

        Examples:

            Create new notifications:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import TransformationNotification
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> notifications = [TransformationNotification(transformation_id = 1, destination="my@email.com"), TransformationNotification(transformation_external_id="transformation2", destination="other@email.com"))]
                >>> res = client.transformations.notifications.create(notifications)
        """
        return run_sync(self.__async_client.transformations.notifications.create(notification=notification))

    def list(
        self,
        transformation_id: int | None = None,
        transformation_external_id: str | None = None,
        destination: str | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
    ) -> TransformationNotificationList:
        """
        `List notification subscriptions. <https://developer.cognite.com/api#tag/Transformation-Notifications/operation/getTransformationNotifications>`_

        Args:
            transformation_id (int | None): Filter by transformation internal numeric ID.
            transformation_external_id (str | None): Filter by transformation externalId.
            destination (str | None): Filter by notification destination.
            limit (int | None): Limits the number of results to be returned. To retrieve all results use limit=-1, default limit is 25.

        Returns:
            TransformationNotificationList: List of transformation notifications

        Example:

            List all notifications::

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> notifications_list = client.transformations.notifications.list()

            List all notifications by transformation id::

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> notifications_list = client.transformations.notifications.list(transformation_id = 1)
        """
        return run_sync(
            self.__async_client.transformations.notifications.list(
                transformation_id=transformation_id,
                transformation_external_id=transformation_external_id,
                destination=destination,
                limit=limit,
            )
        )

    def delete(self, id: int | Sequence[int] | None = None) -> None:
        """
        `Deletes the specified notification subscriptions on the transformation. Does nothing when the subscriptions already don't exist <https://developer.cognite.com/api#tag/Transformation-Notifications/operation/deleteTransformationNotifications>`_

        Args:
            id (int | Sequence[int] | None): Id or list of transformation notification ids

        Examples:

            Delete schedules by id or external id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> client.transformations.notifications.delete(id=[1,2,3])
        """
        return run_sync(self.__async_client.transformations.notifications.delete(id=id))



from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from typing_extensions import Self

from cognite.client.data_classes._base import CogniteObject, CogniteResource, CogniteResourceList
from cognite.client.utils._text import convert_all_keys_to_camel_case

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


@dataclass
class InstanceStatistics(CogniteObject):
    """Statistics for instances in the data modeling API.

    Attributes:
        edges (int): Number of edges in the project.
        soft_deleted_edges (int): Number of soft-deleted edges in the project.
        nodes (int): Number of nodes in the project.
        soft_deleted_nodes (int): Number of soft-deleted nodes in the project.
        instances (int): Total number of instances in the project.
        instances_limit (int): Maximum number of instances allowed in the project.
        soft_deleted_instances (int): Total number of soft-deleted instances in the project.
        soft_deleted_instances_limit (int): Maximum number of soft-deleted instances allowed in the project.
    """

    edges: int
    soft_deleted_edges: int
    nodes: int
    soft_deleted_nodes: int
    instances: int
    instances_limit: int
    soft_deleted_instances: int
    soft_deleted_instances_limit: int

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: "AsyncCogniteClient | None" = None) -> Self:
        return cls(
            edges=resource["edges"],
            soft_deleted_edges=resource["softDeletedEdges"],
            nodes=resource["nodes"],
            soft_deleted_nodes=resource["softDeletedNodes"],
            instances=resource["instances"],
            instances_limit=resource["instancesLimit"],
            soft_deleted_instances=resource["softDeletedInstances"],
            soft_deleted_instances_limit=resource["softDeletedInstancesLimit"],
        )


@dataclass
class CountLimit(CogniteObject):
    """Usage and limits for a specific resource in the data modeling API.

    Attributes:
        count (int): The current usage count for the resource.
        limit (int): The maximum allowed limit for the resource.

    """

    count: int
    limit: int

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: "AsyncCogniteClient | None" = None) -> Self:
        return cls(count=resource["count"], limit=resource["limit"])


@dataclass
class SpaceStatistics(CogniteResource):
    """Statistics for a space in the data modeling API.

    Attributes:
        space (str): The space name
        containers (int): Number of containers in the space.
        views (int): Number of views in the space.
        data_models (int): Number of data models in the space.
        nodes (int): Number of nodes in the space.
        edges (int): Number of edges in the space.
        soft_deleted_nodes (int): Number of soft-deleted nodes in the space.
        soft_deleted_edges (int): Number of soft-deleted edges in the space.

    """

    space: str
    containers: int
    views: int
    data_models: int
    edges: int
    soft_deleted_edges: int
    nodes: int
    soft_deleted_nodes: int

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: "AsyncCogniteClient | None" = None) -> Self:
        instance = cls(
            space=resource["space"],
            containers=resource["containers"],
            views=resource["views"],
            data_models=resource["dataModels"],
            edges=resource["edges"],
            soft_deleted_edges=resource["softDeletedEdges"],
            nodes=resource["nodes"],
            soft_deleted_nodes=resource["softDeletedNodes"],
        )
        instance._cognite_client = cognite_client
        return instance


@dataclass
class ProjectStatistics(CogniteResource):
    """Statistics for a project in the data modeling API.

    Attributes:
        spaces (CountLimit): Usage and limits for spaces in the project
        containers (CountLimit): Usage and limits for containers in the project
        views (CountLimit): Usage and limits for views including all versions in the project
        data_models (CountLimit): Usage and limits for data models including all versions in the project
        container_properties (CountLimit): Usage and limits for sum of container properties in the project
        instances (InstanceStatistics): Usage and limits for number of instances in the project
        concurrent_read_limit (int): Maximum number of concurrent read operations allowed in the project
        concurrent_write_limit (int): Maximum number of concurrent write operations allowed in the project
        concurrent_delete_limit (int): Maximum number of concurrent delete operations allowed in the project
    """

    spaces: CountLimit
    containers: CountLimit
    views: CountLimit
    data_models: CountLimit
    container_properties: CountLimit
    instances: InstanceStatistics
    concurrent_read_limit: int
    concurrent_write_limit: int
    concurrent_delete_limit: int

    @property
    def project(self) -> str:
        """The project name."""
        return self._cognite_client.config.project

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: "AsyncCogniteClient | None" = None) -> Self:
        instance = cls(
            spaces=CountLimit._load(resource["spaces"]),
            containers=CountLimit._load(resource["containers"]),
            views=CountLimit._load(resource["views"]),
            data_models=CountLimit._load(resource["dataModels"]),
            container_properties=CountLimit._load(resource["containerProperties"]),
            instances=InstanceStatistics._load(resource["instances"]),
            concurrent_read_limit=resource["concurrentReadLimit"],
            concurrent_write_limit=resource["concurrentWriteLimit"],
            concurrent_delete_limit=resource["concurrentDeleteLimit"],
        )
        instance._cognite_client = cognite_client
        return instance

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """Dump the object to a dictionary."""
        dumped = {
            "spaces": self.spaces.dump(camel_case),
            "containers": self.containers.dump(camel_case),
            "views": self.views.dump(camel_case),
            "data_models": self.data_models.dump(camel_case),
            "container_properties": self.container_properties.dump(camel_case),
            "instances": self.instances.dump(camel_case),
            "concurrent_read_limit": self.concurrent_read_limit,
            "concurrent_write_limit": self.concurrent_write_limit,
            "concurrent_delete_limit": self.concurrent_delete_limit,
        }
        if camel_case:
            return convert_all_keys_to_camel_case(dumped)
        return dumped


class SpaceStatisticsList(CogniteResourceList[SpaceStatistics]):
    _RESOURCE = SpaceStatistics



"""
===============================================================================
a80fccb60645595b2eb2e245964f25fb
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from collections.abc import Sequence

from cognite.client import AsyncCogniteClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes import (
    TransformationJob,
    TransformationJobList,
    TransformationJobMetricList,
)
from cognite.client.utils._async_helpers import run_sync


class SyncTransformationJobsAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client

    def list(
        self,
        limit: int | None = DEFAULT_LIMIT_READ,
        transformation_id: int | None = None,
        transformation_external_id: str | None = None,
    ) -> TransformationJobList:
        """
        `List all running transformation jobs. <https://developer.cognite.com/api#tag/Transformation-Jobs/operation/getTransformationJobs>`_

        Args:
            limit (int | None): Limits the number of results to be returned. To retrieve all results use limit=-1, default limit is 25.
            transformation_id (int | None): Filters the results by the internal transformation id.
            transformation_external_id (str | None): Filters the results by the external transformation id.

        Returns:
            TransformationJobList: List of transformation jobs

        Example:

            List transformation jobs::

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> transformation_jobs_list = client.transformations.jobs.list()

            List transformation jobs of a single transformation::

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> transformation_jobs_list = client.transformations.jobs.list(transformation_id=1)
        """
        return run_sync(
            self.__async_client.transformations.jobs.list(
                limit=limit, transformation_id=transformation_id, transformation_external_id=transformation_external_id
            )
        )

    def retrieve(self, id: int) -> TransformationJob | None:
        """
        `Retrieve a single transformation job by id. <https://developer.cognite.com/api#tag/Transformation-Jobs/operation/getTransformationJobsByIds>`_

        Args:
            id (int): Job internal Id

        Returns:
            TransformationJob | None: Requested transformation job or None if it does not exist.

        Examples:

            Get transformation job by id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.transformations.jobs.retrieve(id=1)
        """
        return run_sync(self.__async_client.transformations.jobs.retrieve(id=id))

    def list_metrics(self, id: int) -> TransformationJobMetricList:
        """
        `List the metrics of a single transformation job. <https://developer.cognite.com/api#tag/Transformation-Jobs/operation/getTransformationJobsMetrics>`_

        Args:
            id (int): Job internal Id

        Returns:
            TransformationJobMetricList: List of updated metrics of the given job.

        Examples:

            Get metrics by transformation job id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.transformations.jobs.list_metrics(id=1)
        """
        return run_sync(self.__async_client.transformations.jobs.list_metrics(id=id))

    def retrieve_multiple(self, ids: Sequence[int], ignore_unknown_ids: bool = False) -> TransformationJobList:
        """
        `Retrieve multiple transformation jobs by id. <https://developer.cognite.com/api#tag/Transformation-Jobs/operation/getTransformationJobsByIds>`_

        Args:
            ids (Sequence[int]): Job internal Ids
            ignore_unknown_ids (bool): Ignore IDs that are not found rather than throw an exception.

        Returns:
            TransformationJobList: Requested transformation jobs.

        Examples:

            Get jobs by id:

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.transformations.jobs.retrieve_multiple(ids=[1, 2, 3])
        """
        return run_sync(
            self.__async_client.transformations.jobs.retrieve_multiple(ids=ids, ignore_unknown_ids=ignore_unknown_ids)
        )



"""
===============================================================================
215983ac4df951f01a198f0eca4a529a
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, overload

from cognite.client import AsyncCogniteClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes.simulators.simulators import SimulatorIntegration, SimulatorIntegrationList
from cognite.client.utils._async_helpers import SyncIterator, run_sync
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


class SyncSimulatorIntegrationsAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client

    @overload
    def __call__(
        self,
        chunk_size: int,
        simulator_external_ids: str | SequenceNotStr[str] | None = None,
        active: bool | None = None,
        limit: int | None = None,
    ) -> Iterator[SimulatorIntegrationList]: ...

    @overload
    def __call__(
        self,
        chunk_size: None = None,
        simulator_external_ids: str | SequenceNotStr[str] | None = None,
        active: bool | None = None,
        limit: int | None = None,
    ) -> Iterator[SimulatorIntegration]: ...

    def __call__(
        self,
        chunk_size: int | None = None,
        simulator_external_ids: str | SequenceNotStr[str] | None = None,
        active: bool | None = None,
        limit: int | None = None,
    ) -> Iterator[SimulatorIntegration] | Iterator[SimulatorIntegrationList]:
        """
        Iterate over simulator integrations

        Fetches simulator integrations as they are iterated over, so you keep a limited number of simulator integrations in memory.

        Args:
            chunk_size (int | None): Number of simulator integrations to return in each chunk. Defaults to yielding one simulator integration a time.
            simulator_external_ids (str | SequenceNotStr[str] | None): Filter on simulator external ids.
            active (bool | None): Filter on active status of the simulator integration.
            limit (int | None): The maximum number of simulator integrations to return, pass None to return all.

        Yields:
            SimulatorIntegration | SimulatorIntegrationList: yields SimulatorIntegration one by one if chunk_size is not specified, else SimulatorIntegrationList objects.
        """  # noqa: DOC404
        yield from SyncIterator(
            self.__async_client.simulators.integrations(
                chunk_size=chunk_size, simulator_external_ids=simulator_external_ids, active=active, limit=limit
            )
        )  # type: ignore [misc]

    def list(
        self,
        limit: int | None = DEFAULT_LIMIT_READ,
        simulator_external_ids: str | SequenceNotStr[str] | None = None,
        active: bool | None = None,
    ) -> SimulatorIntegrationList:
        """
        `Filter simulator integrations <https://developer.cognite.com/api#tag/Simulator-Integrations/operation/filter_simulator_integrations_simulators_integrations_list_post>`_

        Retrieves a list of simulator integrations that match the given criteria.

        Args:
            limit (int | None): The maximum number of simulator integrations to return, pass None to return all.
            simulator_external_ids (str | SequenceNotStr[str] | None): Filter on simulator external ids.
            active (bool | None): Filter on active status of the simulator integration.

        Returns:
            SimulatorIntegrationList: List of simulator integrations

        Examples:
            List a few simulator integrations:
                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.simulators.integrations.list(limit=10)

            Iterate over simulator integrations, one-by-one:
                >>> for integration in client.simulators.integrations():
                ...     integration  # do something with the simulator integration

            Filter simulator integrations by simulator external ids and active status:
                >>> res = client.simulators.integrations.list(
                ...     simulator_external_ids=["sim1", "sim2"],
                ...     active=True,
                ... )
        """
        return run_sync(
            self.__async_client.simulators.integrations.list(
                limit=limit, simulator_external_ids=simulator_external_ids, active=active
            )
        )

    def delete(
        self, ids: int | Sequence[int] | None = None, external_ids: str | SequenceNotStr[str] | None = None
    ) -> None:
        """
        `Delete simulator integrations <https://developer.cognite.com/api#tag/Simulator-Integrations/operation/delete_simulator_integrations_simulators_integrations_delete_post>`_

        Args:
            ids (int | Sequence[int] | None): Id(s) of simulator integrations to delete
            external_ids (str | SequenceNotStr[str] | None): External_id(s) of simulator integrations to delete

        Examples:
            Delete simulator integrations by id or external id:
                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> client.simulators.integrations.delete(ids=[1,2,3], external_ids="foo")
        """
        return run_sync(self.__async_client.simulators.integrations.delete(ids=ids, external_ids=external_ids))



"""
===============================================================================
d0ad298bed6251d4f457a2a4c2da0e8d
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, overload

from cognite.client import AsyncCogniteClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes.shared import TimestampRange
from cognite.client.data_classes.simulators.filters import SimulationRunsSort
from cognite.client.data_classes.simulators.runs import (
    SimulationRun,
    SimulationRunDataList,
    SimulationRunList,
    SimulationRunWrite,
)
from cognite.client.utils._async_helpers import SyncIterator, run_sync
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


class SyncSimulatorRunsAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client

    @overload
    def __call__(
        self,
        chunk_size: int,
        limit: int | None = None,
        status: str | None = None,
        run_type: str | None = None,
        model_external_ids: SequenceNotStr[str] | None = None,
        simulator_integration_external_ids: SequenceNotStr[str] | None = None,
        simulator_external_ids: SequenceNotStr[str] | None = None,
        routine_external_ids: SequenceNotStr[str] | None = None,
        routine_revision_external_ids: SequenceNotStr[str] | None = None,
        model_revision_external_ids: SequenceNotStr[str] | None = None,
        created_time: TimestampRange | None = None,
        simulation_time: TimestampRange | None = None,
        sort: SimulationRunsSort | None = None,
    ) -> Iterator[SimulationRunList]: ...

    @overload
    def __call__(
        self,
        chunk_size: None = None,
        limit: int | None = None,
        status: str | None = None,
        run_type: str | None = None,
        model_external_ids: SequenceNotStr[str] | None = None,
        simulator_integration_external_ids: SequenceNotStr[str] | None = None,
        simulator_external_ids: SequenceNotStr[str] | None = None,
        routine_external_ids: SequenceNotStr[str] | None = None,
        routine_revision_external_ids: SequenceNotStr[str] | None = None,
        model_revision_external_ids: SequenceNotStr[str] | None = None,
        created_time: TimestampRange | None = None,
        simulation_time: TimestampRange | None = None,
        sort: SimulationRunsSort | None = None,
    ) -> Iterator[SimulationRun]: ...

    def __call__(
        self,
        chunk_size: int | None = None,
        limit: int | None = None,
        status: str | None = None,
        run_type: str | None = None,
        model_external_ids: SequenceNotStr[str] | None = None,
        simulator_integration_external_ids: SequenceNotStr[str] | None = None,
        simulator_external_ids: SequenceNotStr[str] | None = None,
        routine_external_ids: SequenceNotStr[str] | None = None,
        routine_revision_external_ids: SequenceNotStr[str] | None = None,
        model_revision_external_ids: SequenceNotStr[str] | None = None,
        created_time: TimestampRange | None = None,
        simulation_time: TimestampRange | None = None,
        sort: SimulationRunsSort | None = None,
    ) -> Iterator[SimulationRun] | Iterator[SimulationRunList]:
        """
        Iterate over simulation runs

        Fetches simulation runs as they are iterated over, so you keep a limited number of simulation runs in memory.

        Args:
            chunk_size (int | None): Number of simulation runs to return in each chunk. Defaults to yielding one simulation run a time.
            limit (int | None): The maximum number of simulation runs to return, pass None to return all.
            status (str | None): Filter by simulation run status
            run_type (str | None): Filter by simulation run type
            model_external_ids (SequenceNotStr[str] | None): Filter by simulator model external ids
            simulator_integration_external_ids (SequenceNotStr[str] | None): Filter by simulator integration external ids
            simulator_external_ids (SequenceNotStr[str] | None): Filter by simulator external ids
            routine_external_ids (SequenceNotStr[str] | None): Filter by routine external ids
            routine_revision_external_ids (SequenceNotStr[str] | None): Filter by routine revision external ids
            model_revision_external_ids (SequenceNotStr[str] | None): Filter by model revision external ids
            created_time (TimestampRange | None): Filter by created time
            simulation_time (TimestampRange | None): Filter by simulation time
            sort (SimulationRunsSort | None): The criteria to sort by.

        Yields:
            SimulationRun | SimulationRunList: yields Simulation Run one by one if chunk is not specified, else SimulatorRunsList objects.
        """  # noqa: DOC404
        yield from SyncIterator(
            self.__async_client.simulators.runs(
                chunk_size=chunk_size,
                limit=limit,
                status=status,
                run_type=run_type,
                model_external_ids=model_external_ids,
                simulator_integration_external_ids=simulator_integration_external_ids,
                simulator_external_ids=simulator_external_ids,
                routine_external_ids=routine_external_ids,
                routine_revision_external_ids=routine_revision_external_ids,
                model_revision_external_ids=model_revision_external_ids,
                created_time=created_time,
                simulation_time=simulation_time,
                sort=sort,
            )
        )  # type: ignore [misc]

    def list(
        self,
        limit: int | None = DEFAULT_LIMIT_READ,
        status: str | None = None,
        run_type: str | None = None,
        model_external_ids: SequenceNotStr[str] | None = None,
        simulator_integration_external_ids: SequenceNotStr[str] | None = None,
        simulator_external_ids: SequenceNotStr[str] | None = None,
        routine_external_ids: SequenceNotStr[str] | None = None,
        routine_revision_external_ids: SequenceNotStr[str] | None = None,
        model_revision_external_ids: SequenceNotStr[str] | None = None,
        created_time: TimestampRange | None = None,
        simulation_time: TimestampRange | None = None,
        sort: SimulationRunsSort | None = None,
    ) -> SimulationRunList:
        """
        `Filter simulation runs <https://developer.cognite.com/api#tag/Simulation-Runs/operation/filter_simulation_runs_simulators_runs_list_post>`_

        Retrieves a list of simulation runs that match the given criteria.

        Args:
            limit (int | None): The maximum number of simulation runs to return, pass None to return all.
            status (str | None): Filter by simulation run status
            run_type (str | None): Filter by simulation run type
            model_external_ids (SequenceNotStr[str] | None): Filter by simulator model external ids
            simulator_integration_external_ids (SequenceNotStr[str] | None): Filter by simulator integration external ids
            simulator_external_ids (SequenceNotStr[str] | None): Filter by simulator external ids
            routine_external_ids (SequenceNotStr[str] | None): Filter by routine external ids
            routine_revision_external_ids (SequenceNotStr[str] | None): Filter by routine revision external ids
            model_revision_external_ids (SequenceNotStr[str] | None): Filter by model revision external ids
            created_time (TimestampRange | None): Filter by created time
            simulation_time (TimestampRange | None): Filter by simulation time
            sort (SimulationRunsSort | None): The criteria to sort by.

        Returns:
            SimulationRunList: List of simulation runs

        Examples:
            List simulation runs:
                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.simulators.runs.list()

            Iterate over simulation runs, one-by-one:
                >>> for run in client.simulators.runs():
                ...     run  # do something with the simulation run

            Filter runs by status and simulator external ids:
                >>> res = client.simulators.runs.list(
                ...     simulator_external_ids=["PROSPER", "DWSIM"],
                ...     status="success"
                ... )

            Filter runs by time ranges:
                >>> from cognite.client.data_classes.shared import TimestampRange
                >>> res = client.simulators.runs.list(
                ...     created_time=TimestampRange(min=0, max=1_700_000_000_000),
                ...     simulation_time=TimestampRange(min=0, max=1_700_000_000_000),
                ... )
        """
        return run_sync(
            self.__async_client.simulators.runs.list(
                limit=limit,
                status=status,
                run_type=run_type,
                model_external_ids=model_external_ids,
                simulator_integration_external_ids=simulator_integration_external_ids,
                simulator_external_ids=simulator_external_ids,
                routine_external_ids=routine_external_ids,
                routine_revision_external_ids=routine_revision_external_ids,
                model_revision_external_ids=model_revision_external_ids,
                created_time=created_time,
                simulation_time=simulation_time,
                sort=sort,
            )
        )

    @overload
    def retrieve(self, ids: int) -> SimulationRun | None: ...

    @overload
    def retrieve(self, ids: Sequence[int]) -> SimulationRunList | None: ...

    def retrieve(self, ids: int | Sequence[int]) -> SimulationRun | SimulationRunList | None:
        """
        `Retrieve simulation runs by ID <https://api-docs.cognite.com/20230101/tag/Simulation-Runs/operation/simulation_by_id_simulators_runs_byids_post>`_

        Args:
            ids (int | Sequence[int]): The ID(s) of the simulation run(s) to retrieve.

        Returns:
            SimulationRun | SimulationRunList | None: The simulation run(s) with the given ID(s)

        Examples:
            Retrieve a single simulation run by id:
                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> run = client.simulators.runs.retrieve(ids=2)
        """
        return run_sync(self.__async_client.simulators.runs.retrieve(ids=ids))

    @overload
    def create(self, items: SimulationRunWrite) -> SimulationRun: ...

    @overload
    def create(self, items: Sequence[SimulationRunWrite]) -> SimulationRunList: ...

    def create(self, items: SimulationRunWrite | Sequence[SimulationRunWrite]) -> SimulationRun | SimulationRunList:
        """
        `Create simulation runs <https://developer.cognite.com/api#tag/Simulation-Runs/operation/run_simulation_simulators_run_post>`_

        Args:
            items (SimulationRunWrite | Sequence[SimulationRunWrite]): The simulation run(s) to execute.

        Returns:
            SimulationRun | SimulationRunList: Created simulation run(s)

        Examples:
            Create new simulation run:
                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.simulators.runs import SimulationRunWrite
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> run = [
                ...     SimulationRunWrite(
                ...         routine_external_id="routine1",
                ...         log_severity="Debug",
                ...         run_type="external",
                ...     ),
                ... ]
                >>> res = client.simulators.runs.create(run)
        """
        return run_sync(self.__async_client.simulators.runs.create(items=items))

    def list_run_data(self, run_id: int) -> SimulationRunDataList:
        """
        `Get simulation run data <https://developer.cognite.com/api#tag/Simulation-Runs/operation/simulation_data_by_run_id_simulators_runs_data_list_post>`_

        Retrieve data associated with a simulation run by ID.

        Args:
            run_id (int): Simulation run id.

        Returns:
            SimulationRunDataList: List of simulation run data

        Examples:
            Get simulation run data by run id:
                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.simulators.runs.list_run_data(run_id=12345)

            Get simulation run data directly on a simulation run object:
                >>> run = client.simulators.runs.retrieve(ids=2)
                >>> res = run.get_data()
        """
        return run_sync(self.__async_client.simulators.runs.list_run_data(run_id=run_id))



from __future__ import annotations

from collections.abc import AsyncIterator
from typing import TYPE_CHECKING, overload

from cognite.client._api.simulators.integrations import SimulatorIntegrationsAPI
from cognite.client._api.simulators.logs import SimulatorLogsAPI
from cognite.client._api.simulators.models import SimulatorModelsAPI
from cognite.client._api.simulators.routines import SimulatorRoutinesAPI
from cognite.client._api.simulators.runs import SimulatorRunsAPI
from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.simulators.simulators import Simulator, SimulatorList
from cognite.client.utils._experimental import FeaturePreviewWarning

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient
    from cognite.client.config import ClientConfig


class SimulatorsAPI(APIClient):
    _RESOURCE_PATH = "/simulators"

    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: AsyncCogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self.integrations = SimulatorIntegrationsAPI(config, api_version, cognite_client)
        self.models = SimulatorModelsAPI(config, api_version, cognite_client)
        self.runs = SimulatorRunsAPI(config, api_version, cognite_client)
        self.routines = SimulatorRoutinesAPI(config, api_version, cognite_client)
        self.logs = SimulatorLogsAPI(config, api_version, cognite_client)
        self._warning = FeaturePreviewWarning(
            api_maturity="General Availability", sdk_maturity="alpha", feature_name="Simulators"
        )

    @overload
    def __call__(self, chunk_size: None = None, limit: int | None = None) -> AsyncIterator[Simulator]: ...

    @overload
    def __call__(self, chunk_size: int, limit: int | None = None) -> AsyncIterator[SimulatorList]: ...

    async def __call__(
        self, chunk_size: int | None = None, limit: int | None = None
    ) -> AsyncIterator[Simulator] | AsyncIterator[SimulatorList]:
        """Iterate over simulators

        Fetches simulators as they are iterated over, so you keep a limited number of simulators in memory.

        Args:
            chunk_size (int | None): Number of simulators to return in each chunk. Defaults to yielding one simulator a time.
            limit (int | None): Maximum number of simulators to return. Defaults to return all items.

        Yields:
            Simulator | SimulatorList: yields Simulator one by one if chunk is not specified, else SimulatorList objects.
        """  # noqa: DOC404
        async for item in self._list_generator(
            list_cls=SimulatorList,
            resource_cls=Simulator,
            method="POST",
            chunk_size=chunk_size,
            limit=limit,
        ):
            yield item

    async def list(self, limit: int | None = DEFAULT_LIMIT_READ) -> SimulatorList:
        """`List all simulators <https://developer.cognite.com/api#tag/Simulators/operation/filter_simulators_simulators_list_post>`_

        Args:
            limit (int | None): Maximum number of results to return. Defaults to 25. Set to -1, float(“inf”) or None to return all items.

        Returns:
            SimulatorList: List of simulators

        Examples:
            List simulators:
                    >>> from cognite.client import CogniteClient, AsyncCogniteClient
                    >>> client = CogniteClient()
                    >>> # async_client = AsyncCogniteClient()  # another option
                    >>> res = client.simulators.list(limit=10)

            Iterate over simulators, one-by-one:
                    >>> for simulator in client.simulators():
                    ...     simulator  # do something with the simulator

        """
        self._warning.warn()
        return await self._list(method="POST", limit=limit, resource_cls=Simulator, list_cls=SimulatorList)



from __future__ import annotations

import asyncio
import functools
import logging
import random
from collections.abc import (
    AsyncIterable,
    AsyncIterator,
    Callable,
    Coroutine,
    Iterable,
    Mapping,
    MutableMapping,
)
from contextlib import asynccontextmanager
from http.cookiejar import Cookie, CookieJar
from typing import Any, Literal, TypeAlias

import httpx

from cognite.client.config import global_config
from cognite.client.exceptions import (
    CogniteConnectionError,
    CogniteConnectionRefused,
    CogniteHTTPStatusError,
    CogniteReadTimeout,
    CogniteRequestError,
)
from cognite.client.response import CogniteHTTPResponse
from cognite.client.utils._concurrency import get_global_semaphore

logger = logging.getLogger(__name__)


HTTPResponseCoro: TypeAlias = Coroutine[Any, Any, httpx.Response]


class NoCookiesPlease(CookieJar):
    def set_cookie(self, cookie: Cookie) -> None:
        pass


@functools.cache
def get_global_async_httpx_client() -> httpx.AsyncClient:
    async_transport = httpx.AsyncHTTPTransport(
        proxy=global_config.proxy,
        retries=0,  # 'retries': The maximum number of retries when trying to establish a connection.
        verify=not global_config.disable_ssl,
        limits=httpx.Limits(
            # max_connections: The maximum number of concurrent HTTP connections that
            #     the pool should allow. Any attempt to send a request on a pool that
            #     would exceed this amount will block until a connection is available.
            # max_keepalive_connections: The maximum number of idle HTTP connections
            #     that will be maintained in the pool.
            # keepalive_expiry: The duration in seconds that an idle HTTP connection
            #     may be maintained for before being expired from the pool.
            max_connections=global_config.max_connection_pool_size,
            max_keepalive_connections=None,  # defaults to match max_connections
            keepalive_expiry=5,  # copy httpx default
        ),
    )
    return httpx.AsyncClient(
        transport=async_transport,
        follow_redirects=global_config.follow_redirects,
        cookies=NoCookiesPlease(),
        # Below should not be needed when we pass transport, but... :)
        proxy=global_config.proxy,
        verify=not global_config.disable_ssl,
    )


class AsyncHTTPClientWithRetryConfig:
    def __init__(
        self,
        status_codes_to_retry: set[int] | None = None,
        backoff_factor: float = 0.5,
        max_backoff_seconds: int | None = None,
        max_retries_total: int | None = None,
        max_retries_status: int | None = None,
        max_retries_read: int | None = None,
        max_retries_connect: int | None = None,
    ) -> None:
        self._status_codes_to_retry = status_codes_to_retry
        self.backoff_factor = backoff_factor
        self._max_backoff_seconds = max_backoff_seconds
        self._max_retries_total = max_retries_total
        self._max_retries_status = max_retries_status
        self._max_retries_read = max_retries_read
        self._max_retries_connect = max_retries_connect

    @property
    def status_codes_to_retry(self) -> set[int]:
        if self._status_codes_to_retry is None:
            # Changes to the global config need to take effect immediately
            return global_config.status_forcelist
        return self._status_codes_to_retry

    @property
    def max_backoff_seconds(self) -> int:
        if self._max_backoff_seconds is None:
            return global_config.max_retry_backoff
        return self._max_backoff_seconds

    @property
    def max_retries_total(self) -> int:
        if self._max_retries_total is None:
            return global_config.max_retries
        return self._max_retries_total

    @property
    def max_retries_status(self) -> int:
        if self._max_retries_status is None:
            return global_config.max_retries
        return self._max_retries_status

    @property
    def max_retries_read(self) -> int:
        if self._max_retries_read is None:
            return global_config.max_retries
        return self._max_retries_read

    @property
    def max_retries_connect(self) -> int:
        if self._max_retries_connect is None:
            return global_config.max_retries_connect
        return self._max_retries_connect


class RetryTracker:
    def __init__(self, url: str, config: AsyncHTTPClientWithRetryConfig) -> None:
        self.url = url
        self.config = config
        self.status = self.read = self.connect = 0
        self.last_failed_reason = ""
        self.total_time_in_backoff = 0.0

    @property
    def total(self) -> int:
        return self.status + self.read + self.connect

    def get_backoff_time(self) -> float:
        """
        Computes the randomized delay (backoff time) before the next retry attempt.

        This method implements the **Exponential Backoff with Full Jitter** strategy,
        with one addition, a small "initial floor" to avoid immediate retries.
        """
        base = self.config.backoff_factor * 2**self.total
        cap = min(base, self.config.max_backoff_seconds)
        return 0.1 + random.uniform(0, cap)

    async def back_off(self) -> None:
        backoff_time = self.get_backoff_time()
        self.total_time_in_backoff += backoff_time
        # We put logging here, since this is always called before retrying
        logger.debug(
            f"Retrying failed request, attempt #{self.total}, backoff wait: {backoff_time:.4f} sec "
            f"(total: {self.total_time_in_backoff:.4f} sec), reason: {self.last_failed_reason!r}, url: {self.url}"
        )
        await asyncio.sleep(backoff_time)

    @property
    def should_retry_total(self) -> bool:
        # We use 'less than or equal' because should_retry_total is always checked after we have bumped
        # one of [status, read, connect] += 1. Said differently, do last retry when 'total = max':
        return self.total <= self.config.max_retries_total

    def should_retry_status_code(self, err: httpx.HTTPStatusError, is_auto_retryable: bool = False) -> bool:
        self.status += 1
        status_code = err.response.status_code
        error_type = CogniteHTTPStatusError.get_error_type(status_code)
        self.last_failed_reason = f"HTTPStatusError({status_code} - {error_type}: {err.response.reason_phrase})"
        return (
            self.should_retry_total
            and self.status <= self.config.max_retries_status
            and (is_auto_retryable or status_code in self.config.status_codes_to_retry)
        )

    def should_retry_connect_error(self, error: httpx.TransportError | httpx.DecodingError) -> bool:
        self.connect += 1
        self.last_failed_reason = f"{type(error).__name__}({error!s})"
        return self.should_retry_total and self.connect <= self.config.max_retries_connect

    def should_retry_timeout(self, error: httpx.TimeoutException) -> bool:
        self.read += 1
        self.last_failed_reason = f"{type(error).__name__}({error!s})"
        return self.should_retry_total and self.read <= self.config.max_retries_read


class AsyncHTTPClientWithRetry:
    def __init__(
        self,
        config: AsyncHTTPClientWithRetryConfig,
        refresh_auth_header: Callable[[MutableMapping[str, str]], None],
        httpx_async_client: httpx.AsyncClient | None = None,
    ) -> None:
        self.config = config
        self.refresh_auth_header = refresh_auth_header
        self.httpx_async_client = httpx_async_client or get_global_async_httpx_client()

    async def request(
        self,
        method: Literal["GET", "POST", "PUT", "PATCH", "DELETE", "HEAD", "OPTIONS"],
        /,
        url: str,
        *,
        content: str | bytes | Iterable[bytes] | AsyncIterable[bytes] | None = None,
        data: Mapping[str, Any] | None = None,
        json: Any = None,
        params: Mapping[str, str] | None = None,
        headers: MutableMapping[str, str] | None = None,
        follow_redirects: bool = False,
        timeout: float | None = None,
        semaphore: asyncio.BoundedSemaphore | None = None,
    ) -> CogniteHTTPResponse:
        def coro_factory() -> HTTPResponseCoro:
            return self.httpx_async_client.request(
                method,
                url,
                content=content,
                data=data,
                json=json,
                params=params,
                headers=headers,
                follow_redirects=follow_redirects,
                timeout=timeout,
            )

        return await self._with_retry(coro_factory, url=url, headers=headers, semaphore=semaphore)

    @asynccontextmanager
    async def stream(
        self,
        method: Literal["GET", "POST"],
        /,
        url: str,
        *,
        json: Any = None,
        headers: MutableMapping[str, str] | None = None,
        timeout: float | None = None,
    ) -> AsyncIterator[CogniteHTTPResponse]:
        # This method is basically a clone of httpx.AsyncClient.stream() so that we may add our own retry logic.
        def coro_factory() -> HTTPResponseCoro:
            request = self.httpx_async_client.build_request(
                method=method, url=url, json=json, headers=headers, timeout=timeout
            )
            return self.httpx_async_client.send(request, stream=True)

        response: CogniteHTTPResponse | None = None
        try:
            yield (response := await self._with_retry(coro_factory, url=url, headers=headers))
        finally:
            if response:
                await response.aclose()

    async def _with_retry(
        self,
        coro_factory: Callable[[], HTTPResponseCoro],
        *,
        url: str,
        headers: MutableMapping[str, str] | None,
        semaphore: asyncio.BoundedSemaphore | None = None,
    ) -> CogniteHTTPResponse:
        if semaphore is None:
            # By default, we run with a semaphore decided by user settings of 'max_workers' in 'global_config'.
            # Since the user can run any number of SDK tasks concurrently, this needs to be global:
            semaphore = get_global_semaphore()

        is_auto_retryable = False
        retry_tracker = RetryTracker(url, self.config)
        accepts_json = (headers or {}).get("accept") == "application/json"
        while True:
            try:
                async with semaphore:
                    # Ensure our credentials are not about to expire right before making the request:
                    if headers is not None:
                        self.refresh_auth_header(headers)
                    response = await coro_factory()

                if accepts_json:
                    # Cache .json() return value in order to avoid redecoding JSON if called multiple times
                    response.json = functools.cache(response.json)  # type: ignore [method-assign]
                return CogniteHTTPResponse(response.raise_for_status())

            except httpx.HTTPStatusError as err:
                response = err.response
                is_auto_retryable = response.headers.get("cdf-is-auto-retryable", False)
                if not retry_tracker.should_retry_status_code(err, is_auto_retryable):
                    raise CogniteHTTPStatusError(
                        response.status_code,
                        request=err.request,
                        response=CogniteHTTPResponse(response),
                    ) from None

            except httpx.ConnectError as err:
                if not retry_tracker.should_retry_connect_error(err):
                    raise CogniteConnectionRefused from err

            except (httpx.NetworkError, httpx.ConnectTimeout, httpx.DecodingError) as err:
                if not retry_tracker.should_retry_connect_error(err):
                    raise CogniteConnectionError from err

            except httpx.TimeoutException as err:
                if not retry_tracker.should_retry_timeout(err):
                    raise CogniteReadTimeout from err

            except httpx.RequestError as err:
                # We want to avoid raising a non-Cognite error (from the underlying library). httpx.RequestError is the
                # base class for all exceptions that can be raised during a request, so we use it here as a fallback.
                raise CogniteRequestError from err

            # If we got here, it means we have decided to retry the request!
            await retry_tracker.back_off()



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal, NoReturn

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteObject,
    CognitePrimitiveUpdate,
    CogniteResource,
    CogniteResourceList,
    CogniteUpdate,
    PropertySpec,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient


@dataclass
class SessionCredentials(CogniteObject):
    nonce: str

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            nonce=resource["nonce"],
        )


class _UserCore(WriteableCogniteResource["UserWrite"], ABC): ...


class UserWrite(_UserCore):
    """A postgres gateway **user** (also a typical postgres user) owns the foreign tables (built in or custom).

    The created postgres user only has access to use foreign tables and cannot directly create tables users. To create
    foreign tables use the Postgres Gateway Tables APIs

    This is the write/request format of the user.

    Args:
        credentials (SessionCredentials | None): Credentials for authenticating towards CDF using a CDF session.

    """

    def __init__(self, credentials: SessionCredentials | None = None) -> None:
        self.credentials = credentials

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            credentials=SessionCredentials._load(resource["credentials"], cognite_client)
            if "credentials" in resource
            else None,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if isinstance(self.credentials, SessionCredentials):
            output["credentials"] = self.credentials.dump(camel_case=camel_case)

        return output

    def as_write(self) -> UserWrite:
        return self


class User(_UserCore):
    """A user.

    This is the read/response format of the user for list and retrieve endpoints.

    Args:
        username (str): Username to authenticate the user on the DB.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        session_id (int | None): ID of the session tied to this user.

    """

    def __init__(
        self,
        username: str,
        created_time: int,
        last_updated_time: int,
        session_id: int | None = None,
    ) -> None:
        self.username = username
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self.session_id = session_id

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            username=resource["username"],
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
            session_id=resource.get("sessionId"),
        )

    def as_write(self) -> NoReturn:
        raise TypeError(f"{type(self).__name__} cannot be converted to a write object")


class UserCreated(User):
    """A user.

    This is the read/response format of the user for the create endpoint

    Args:
        host(str): Host of the DB.
        username (str): Username to authenticate the user on the DB.
        password (str): Password to authenticate the user on the DB.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        session_id (int | None): ID of the session tied to this user.

    """

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
        created_time: int,
        last_updated_time: int,
        session_id: int | None = None,
    ) -> None:
        super().__init__(username, created_time, last_updated_time, session_id)
        self.host = host
        self.password = password

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: AsyncCogniteClient | None = None) -> Self:
        return cls(
            host=resource["host"],
            username=resource["username"],
            password=resource["password"],
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
            session_id=resource.get("sessionId"),
        )


class UserUpdate(CogniteUpdate):
    def __init__(
        self,
        username: str,
    ) -> None:
        super().__init__()
        self.username = username

    class _UpdateItemSessionCredentialsUpdate(CognitePrimitiveUpdate):
        def set(self, value: SessionCredentials | None) -> UserUpdate:
            return self._set(value.dump() if isinstance(value, SessionCredentials) else value)

    @property
    def credentials(self) -> UserUpdate._UpdateItemSessionCredentialsUpdate:
        return self._UpdateItemSessionCredentialsUpdate(self, "credentials")

    @classmethod
    def _get_update_properties(cls, item: CogniteResource | None = None) -> list[PropertySpec]:
        return [
            PropertySpec("credentials", is_nullable=True),
        ]

    def dump(self, camel_case: Literal[True] = True) -> dict[str, Any]:
        """Dump the instance into a json serializable Python data type.

        Args:
            camel_case (Literal[True]): No description.
        Returns:
            dict[str, Any]: A dictionary representation of the instance.
        """
        return {"update": self._update_object, "username": self.username}


class UserWriteList(CogniteResourceList[UserWrite]):
    _RESOURCE = UserWrite


class UserList(WriteableCogniteResourceList[UserWrite, User]):
    _RESOURCE = User

    def as_write(self) -> NoReturn:
        raise TypeError(f"{type(self).__name__} cannot be converted to a write object")


class UserCreatedList(WriteableCogniteResourceList[UserWrite, UserCreated]):
    _RESOURCE = UserCreated

    def as_write(self) -> NoReturn:
        raise TypeError(
            f"{type(self).__name__} cannot be converted to a {UserWrite.__name__} object. "
            f"The {UserWrite.__name__} contains the credentials for the user, which are not returned in the "
            f"response of the create endpoint."
        )



"""
===============================================================================
fea2e0196dbe6a2804d937ee40ff61e9
This file is auto-generated from the Async API modules, - do not edit manually!
===============================================================================
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import AsyncCogniteClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client._sync_api_client import SyncAPIClient
from cognite.client.data_classes import SecurityCategory, SecurityCategoryList
from cognite.client.data_classes.iam import SecurityCategoryWrite
from cognite.client.utils._async_helpers import run_sync


class SyncSecurityCategoriesAPI(SyncAPIClient):
    """Auto-generated, do not modify manually."""

    def __init__(self, async_client: AsyncCogniteClient) -> None:
        self.__async_client = async_client

    def list(self, limit: int | None = DEFAULT_LIMIT_READ) -> SecurityCategoryList:
        """
        `List security categories. <https://developer.cognite.com/api#tag/Security-categories/operation/getSecurityCategories>`_

        Args:
            limit (int | None): Max number of security categories to return. Defaults to 25. Set to -1, float("inf") or None to return all items.

        Returns:
            SecurityCategoryList: List of security categories

        Example:

            List security categories::

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> res = client.iam.security_categories.list()
        """
        return run_sync(self.__async_client.iam.security_categories.list(limit=limit))

    @overload
    def create(self, security_category: SecurityCategory | SecurityCategoryWrite) -> SecurityCategory: ...

    @overload
    def create(
        self, security_category: Sequence[SecurityCategory] | Sequence[SecurityCategoryWrite]
    ) -> SecurityCategoryList: ...

    def create(
        self,
        security_category: SecurityCategory
        | SecurityCategoryWrite
        | Sequence[SecurityCategory]
        | Sequence[SecurityCategoryWrite],
    ) -> SecurityCategory | SecurityCategoryList:
        """
        `Create one or more security categories. <https://developer.cognite.com/api#tag/Security-categories/operation/createSecurityCategories>`_

        Args:
            security_category (SecurityCategory | SecurityCategoryWrite | Sequence[SecurityCategory] | Sequence[SecurityCategoryWrite]): Security category or list of categories to create.

        Returns:
            SecurityCategory | SecurityCategoryList: The created security category or categories.

        Example:

            Create security category::

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import SecurityCategoryWrite
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> my_category = SecurityCategoryWrite(name="My Category")
                >>> res = client.iam.security_categories.create(my_category)
        """
        return run_sync(self.__async_client.iam.security_categories.create(security_category=security_category))

    def delete(self, id: int | Sequence[int]) -> None:
        """
        `Delete one or more security categories. <https://developer.cognite.com/api#tag/Security-categories/operation/deleteSecurityCategories>`_

        Args:
            id (int | Sequence[int]): ID or list of IDs of security categories to delete.

        Example:

            Delete security category::

                >>> from cognite.client import CogniteClient, AsyncCogniteClient
                >>> client = CogniteClient()
                >>> # async_client = AsyncCogniteClient()  # another option
                >>> client.iam.security_categories.delete(1)
        """
        return run_sync(self.__async_client.iam.security_categories.delete(id=id))



from __future__ import annotations

from typing import TYPE_CHECKING, NoReturn, overload

from cognite.client.utils._auxiliary import at_most_one_is_not_none
from cognite.client.utils._identifier import IdentifierSequence

if TYPE_CHECKING:
    from cognite.client import AsyncCogniteClient
    from cognite.client.utils._identifier import Identifier


async def _get_function_internal_id(cognite_client: AsyncCogniteClient, identifier: Identifier) -> int:
    primitive = identifier.as_primitive()
    if identifier.is_id:
        return primitive

    if identifier.is_external_id:
        function = await cognite_client.functions.retrieve(external_id=primitive)
        if function:
            return function.id

    raise ValueError(f'Function with external ID "{primitive}" is not found')


def _get_function_identifier(function_id: int | None, function_external_id: str | None) -> Identifier:
    identifier = IdentifierSequence.load(function_id, function_external_id, id_name="function")
    if identifier.is_singleton():
        return identifier[0]
    raise ValueError("Exactly one of function_id and function_external_id must be specified")


@overload
def _ensure_at_most_one_id_given(function_id: int, function_external_id: str) -> NoReturn: ...


@overload
def _ensure_at_most_one_id_given(function_id: int | None, function_external_id: str | None) -> None: ...


def _ensure_at_most_one_id_given(function_id: int | None, function_external_id: str | None) -> None:
    if at_most_one_is_not_none(function_id, function_external_id):
        return
    raise ValueError("Both 'function_id' and 'function_external_id' were supplied, pass exactly one or neither.")
